/* ************************************************************************
 * Copyright 2006 VMware, Inc.  All rights reserved. -- VMware Confidential
 * ************************************************************************/

/* 
 * vmciDatagram.c --
 *
 *      Simple Datagram API for the Linux guest driver
 */

#if defined(linux)
#include "driver-config.h"
   
#include <linux/kernel.h>  
#include <linux/pci.h>
#elif defined(WIN32)
#include <ntddk.h>
#else
#error "Platform not support by VMCI datagram API."
#endif // linux

#include "vm_basic_types.h"
#include "vm_assert.h"
#include "vmci_defs.h"
#include "vmci_kernel_defs.h"
#include "vmci_infrastructure.h"
#include "vmciInt.h"
#include "vmciUtil.h"
#include "vmciDatagram.h"

typedef struct DatagramHashEntry {
   struct DatagramHashEntry *next;
   int refCount;

   VMCIHandle handle;
   uint32 flags;
   VMCIDatagramCB notifyCB;
   void *notifyData;
} DatagramHashEntry;

#define HASH_TABLE_SIZE 64

/* 
 * Hash table containing all the datagram handles for this VM. It is 
 * synchronized using a single lock but we should consider making it more
 * fine grained, e.g. a per bucket lock or per set of buckets' lock.
 */

typedef struct DatagramHashTable {
   VMCILock lock;
   DatagramHashEntry *entries[HASH_TABLE_SIZE];
} DatagramHashTable; 


static int DatagramHashAddEntry(DatagramHashEntry *entry);
static void DatagramHashRemoveEntry(VMCIHandle handle);
static DatagramHashEntry *DatagramHashGetEntry(VMCIHandle handle);
static void DatagramHashReleaseEntry(DatagramHashEntry *entry);
static Bool DatagramHandleUnique(VMCIHandle handle);
static int DatagramReceive(VMCIId contextID, void *clientData,
                           VMCICall *call);
static int64 DatagramDriverHandler(void *ctObj, 
                                   VMCIObjType ctObjType,
                                   VMCICall *call);
static void DatagramProcessNotify(void *notifyData, uint32 msgSize,
                                  VMCIDatagramMsg *msg);
static int DatagramProcessSendCall(VMCIDatagramProcess *dgmProc,
                                   VMCICall *dgmProcCall,
                                   Bool mustCopy);

DatagramHashTable hashTable;

/*
 *-------------------------------------------------------------------------
 *
 *  DatagramHashAddEntry --
 *
 *  Result:
 *     VMCI_SUCCESS if added, error if not.
 *     
 *-------------------------------------------------------------------------
 */

static int
DatagramHashAddEntry(DatagramHashEntry *entry)
{
   int idx;
   VMCILockFlags flags;

   if (!DatagramHandleUnique(entry->handle)) {
      return VMCI_ERROR_DUPLICATE_ENTRY;
   }
   
   idx = VMCI_Hash(entry->handle, HASH_TABLE_SIZE);

   /* New entry is added to top/front of hash bucket. */
   VMCI_GrabLock_BH(&hashTable.lock, &flags);
   entry->refCount++;
   entry->next = hashTable.entries[idx];
   hashTable.entries[idx] = entry;
   VMCI_ReleaseLock_BH(&hashTable.lock, flags);

   return VMCI_SUCCESS;
}


/*
 *-------------------------------------------------------------------------
 *
 *  DatagramHashRemoveEntry --
 *
 *  Result:
 *     None.
 *     
 *-------------------------------------------------------------------------
 */

static void
DatagramHashRemoveEntry(VMCIHandle handle)
{ 
   VMCILockFlags flags;
   DatagramHashEntry *prev, *cur;
   int idx = VMCI_Hash(handle, HASH_TABLE_SIZE);

   prev = NULL;
   VMCI_GrabLock_BH(&hashTable.lock, &flags);
   cur = hashTable.entries[idx];
   while (TRUE) {
      if (cur == NULL) {
	 break;
      }
      if (cur->handle == handle) {
	 /* Remove entry and break. */
	 if (prev) {
	    prev->next = cur->next;
	 } else {
	    hashTable.entries[idx] = cur->next;
	 }

	 cur->refCount--;
	 
	 /* Check if this is last reference and free if so. */
	 if (cur->refCount == 0) { 
	    VMCI_FreeKernelMem(cur);
	 }
	 break;
      }
      prev = cur;
      cur = cur->next;
   }
   VMCI_ReleaseLock_BH(&hashTable.lock, flags);
}


/*
 *-------------------------------------------------------------------------
 *
 *  DatagramHashGetEntry --
 *
 *  Result:
 *     None.
 *     
 *-------------------------------------------------------------------------
 */

static DatagramHashEntry *
DatagramHashGetEntry(VMCIHandle handle)
{
   VMCILockFlags flags;
   DatagramHashEntry *cur;
   int idx = VMCI_Hash(handle, HASH_TABLE_SIZE);
   
   VMCI_GrabLock_BH(&hashTable.lock, &flags);
   cur = hashTable.entries[idx];
   while (TRUE) {
      if (cur == NULL) {
	 break;
      }

      if (cur->handle == handle) {
	 cur->refCount++;
	 break;
      }
   }
   VMCI_ReleaseLock_BH(&hashTable.lock, flags);
   
   return cur;
}


/*
 *-------------------------------------------------------------------------
 *
 *  DatagramHashReleaseEntry --
 *
 *  Result:
 *     None.
 *     
 *-------------------------------------------------------------------------
 */

static void
DatagramHashReleaseEntry(DatagramHashEntry *entry)
{
   VMCILockFlags flags;

   VMCI_GrabLock_BH(&hashTable.lock, &flags);
   entry->refCount--;

   /* Check if this is last reference and free if so. */
   if (entry->refCount == 0) { 
      VMCI_FreeKernelMem(entry);
   }
   VMCI_ReleaseLock_BH(&hashTable.lock, flags);
}


/*
 *------------------------------------------------------------------------------
 *
 *  DatagramHandleUnique --
 *
 *  Result:
 *     None.
 *     
 *------------------------------------------------------------------------------
 */

static Bool
DatagramHandleUnique(VMCIHandle handle)
{
   VMCILockFlags flags;
   Bool unique = TRUE;
   DatagramHashEntry *entry;
   int idx = VMCI_Hash(handle, HASH_TABLE_SIZE);

   VMCI_GrabLock_BH(&hashTable.lock, &flags);
   entry = hashTable.entries[idx];
   while (entry) {
      if (entry->handle == handle) {
	 unique = FALSE;
	 break;
      }
      entry = entry->next;
   }
   VMCI_ReleaseLock_BH(&hashTable.lock, flags);
   return unique;
}


/*
 *-----------------------------------------------------------------------------
 *
 * VMCIDatagram_CreateHnd --
 *
 *      Creates a datagram endpoint and returns a handle to it.
 *
 * Results:
 *      Returns handle if success, negative errno value otherwise.
 *
 * Side effects:
 *      Datagram endpoint is created both in guest and on host.
 *
 *-----------------------------------------------------------------------------
 */

VMCIHandle
VMCIDatagram_CreateHnd(VMCIId resourceID, // IN:
                       uint32 flags,           // IN:
                       VMCIDatagramCB notifyCB,  // IN:
                       void *notifyData)       // IN:
{
   int result;
   DatagramHashEntry *entry;
   VMCIHandle handle;
   VMCIDatagramCreateHdr createCall;
   VMCIId contextID = VMCI_GetContextID();

   /* Validate contextID. */
   if (contextID == VMCI_INVALID_ID) {
      return VMCI_INVALID_HANDLE;
   }
   
   if (resourceID == VMCI_INVALID_ID) {
      resourceID = VMCI_GetResourceID();
   }

   createCall.hdr.vector = VMCI_DATAGRAM_CREATE_HND;
   createCall.hdr.size = sizeof createCall;
   createCall.resourceID = resourceID;
   createCall.flags = flags;

   handle = VMCI_MAKE_HANDLE(contextID, resourceID);

   /* Update local datastructure. */
   entry = VMCI_AllocKernelMem(sizeof *entry, VMCI_MEMORY_NONPAGED);
   if (entry == NULL) {
      return VMCI_INVALID_HANDLE;
   }
   
   entry->handle = handle;
   entry->flags = flags;
   entry->notifyCB = notifyCB;
   entry->notifyData = notifyData;
   entry->refCount = 0;

   result = DatagramHashAddEntry(entry);
   if (result != VMCI_SUCCESS) {
      VMCI_LOG(("Failed to add new entry, err 0x%x.\n", result));
      VMCI_FreeKernelMem(entry);
      return VMCI_INVALID_HANDLE;
   }

   /* Create the datagram structure on the host. */
   result = VMCI_HyperCall((VMCICall *)&createCall);
   if (result != VMCI_SUCCESS) {
      VMCI_LOG(("Failed to create new datagram handle, err 0x%x.\n",
		     result));
      DatagramHashRemoveEntry(handle);
      return VMCI_INVALID_HANDLE;
   }

   return handle;
}

		      
/*
 *-----------------------------------------------------------------------------
 *
 * VMCIDatagram_DestroyHnd --
 *
 *      Destroys a handle.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      Host and guest state is cleaned up.
 *
 *-----------------------------------------------------------------------------
 */

void
VMCIDatagram_DestroyHnd(VMCIHandle handle)       // IN
{
   VMCIDatagramDestroyHdr destroyCall =
      {{VMCI_DATAGRAM_DESTROY_HND, sizeof destroyCall}, handle};

   /*
    * First destroy handle at host to prevent incoming datagrams after local 
    * struct is removed.
    */
   VMCI_HyperCall((VMCICall *)&destroyCall);
   
   DatagramHashRemoveEntry(handle);
}


/*
 *-----------------------------------------------------------------------------
 *
 * VMCIDatagram_Send --
 *
 *      Sends the payload to the destination datagram handle.
 *
 * Results:
 *      Returns number of bytes sent if success, or error code if failure.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

int
VMCIDatagram_Send(VMCIHandle dstHandle,      // IN
                  VMCIHandle srcHandle,      // IN
                  uint32 payloadSize,      // IN
                  void *payload)           // IN
		
{
   uint32 retval = VMCI_SUCCESS;
   DatagramHashEntry *entry;
   VMCIDatagramSendHdr *sendCall;
   uint32 callSize = sizeof *sendCall + payloadSize;

   /* Check srcHandle exists otherwise fail early. */
   entry = DatagramHashGetEntry(srcHandle);
   if (entry == NULL) {
      VMCI_LOG(("Couldn't find handle 0x%"FMT64"x.\n", srcHandle));
      retval = VMCI_ERROR_INVALID_ARGS;
      goto done;
   }
   
   
   if (callSize > VMCI_CALL_STANDARD_SIZE) {
      VMCI_LOG(("Payload size %d too big to send, max payload is "
		     "%"FMTSZ"d.\n",
		     payloadSize,
		     VMCI_CALL_STANDARD_SIZE - VMCI_CALL_HEADERSIZE));
      retval = VMCI_ERROR_INVALID_ARGS;
      goto release;
   }

   sendCall = VMCI_AllocKernelMem(callSize, 0);
   if (sendCall == NULL) {
      VMCI_LOG(("Couldn't allocate memory for handle 0x%"FMT64"x.\n",
		     srcHandle));
      retval = VMCI_ERROR_NO_MEM;
      goto release;
   }

   sendCall->hdr.vector = VMCI_DATAGRAM_SEND;
   sendCall->hdr.size = callSize;
   sendCall->dstHandle = dstHandle;
   sendCall->srcHandle = srcHandle;
   sendCall->payloadSize = payloadSize;
   if (payloadSize > 0) {
      memcpy((char *)sendCall + sizeof *sendCall, payload, payloadSize);
   }
   retval = VMCI_HyperCall((VMCICall *)sendCall);
   VMCI_FreeKernelMem(sendCall);
   

  release:
   DatagramHashReleaseEntry(entry);

  done:
   return retval;
}


/*
 *-----------------------------------------------------------------------------
 *
 * DatagramReceive --
 *
 *      Forwards the payload to the datagram handle's callback.
 *
 * Results:
 *      VMCI_SUCCESS on success, error code if not.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

static int
DatagramReceive(VMCIId contextID,  // IN: unused always host context ID
                void *clientData,       // IN
                VMCICall *call)    // IN
{
   DatagramHashEntry *entry;
   VMCIDatagramReceiveHdr *receiveCall = (VMCIDatagramReceiveHdr *)call;
   
   if (receiveCall == NULL) {
      VMCI_LOG(("receiveCall is NULL.\n"));
      return VMCI_ERROR_INVALID_ARGS;
   }
   
   if (receiveCall->hdr.size != (sizeof *receiveCall + receiveCall->payloadSize)) {
      VMCI_LOG(("Invalid datagram receive data, call size %d, computed size "
                "%"FMTSZ"d.\n.",
                receiveCall->hdr.size, 
                sizeof *receiveCall + receiveCall->payloadSize));
      return VMCI_ERROR_INVALID_ARGS;
   }

   entry = DatagramHashGetEntry(receiveCall->dstHandle);
   if (entry == NULL) {
      VMCI_LOG(("destination handle 0x%"FMT64"x doesn't exists.\n",
		     receiveCall->dstHandle));
      return VMCI_ERROR_NO_HANDLE;
   }
   
   if (entry->notifyCB) {
      entry->notifyCB(entry->notifyData, 
		      call->h.size - VMCI_CALL_HEADERSIZE,
		      (VMCIDatagramMsg *)call->args);
   } else {
      VMCI_LOG(("no handle callback for handle 0x%"FMT64"x payload of "
		     "size %d.\n", 
		     receiveCall->dstHandle, receiveCall->payloadSize));
   }
   DatagramHashReleaseEntry(entry);

   return VMCI_SUCCESS;
}


/*
 *------------------------------------------------------------------------------
 *
 * DatagramDriverHandler --
 *
 *      Parses the driver call and calls the appropriate function.
 *
 * Results:
 *      Value of called function.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------------------
 */

static int64
DatagramDriverHandler(void *ctObj,         // IN:
                      VMCIObjType ctObjType, // IN:
                      VMCICall *call) // IN:
{
   VMCIDatagramProcess *dgmProc;
   int64 retval;

   if (!ctObj || ctObjType != VMCIOBJ_DATAGRAM_PROCESS || !call) {
      return VMCI_ERROR_INVALID_ARGS;
   }

   dgmProc = (VMCIDatagramProcess *) ctObj;

   switch (call->h.vector) {
   case VMCI_DATAGRAM_CREATE_HND: {
      VMCIHandle handle;
      VMCIDatagramCreateHdr *createCall = (VMCIDatagramCreateHdr *)call;
      if (createCall->hdr.size != sizeof *createCall) {
         retval = VMCI_ERROR_INVALID_ARGS;
         break;
      }
      
      handle = VMCIDatagram_CreateHnd(createCall->resourceID, createCall->flags,
                                      DatagramProcessNotify, (void *) dgmProc);
      
      if (handle != VMCI_INVALID_HANDLE) {
         ASSERT(dgmProc->handle == VMCI_INVALID_HANDLE);
         dgmProc->handle = handle;
      }

      retval = handle;
      break;
   }
   case VMCI_DATAGRAM_DESTROY_HND: {
      VMCIDatagramDestroyHdr *destroyCall = (VMCIDatagramDestroyHdr  *)call;
      
      if (destroyCall->hdr.size != sizeof *destroyCall) {
         retval = VMCI_ERROR_INVALID_ARGS;
         break;
      }

      if (dgmProc->handle != destroyCall->handle) {
         /* not ours */
         retval = VMCI_ERROR_NO_HANDLE;
         break;
      }
      
      VMCIDatagram_DestroyHnd(destroyCall->handle);
      dgmProc->handle = VMCI_INVALID_HANDLE;
      retval = VMCI_SUCCESS;
      break;
   }
   case VMCI_DATAGRAM_SEND: {
      int callSize;
      VMCIDatagramSendHdr *sendCall = (VMCIDatagramSendHdr *)call;
      callSize = sizeof *sendCall + sendCall->payloadSize;
      
      if (call->h.size != callSize) {
         retval = VMCI_ERROR_INVALID_ARGS;
         break;
      }

      /* XXX We could optimize to avoid copy, but must sanity check. */
      retval = VMCIDatagram_Send(sendCall->dstHandle, sendCall->srcHandle,
			       sendCall->payloadSize, 
			       (char *)sendCall + sizeof *sendCall);
      break;
   }
   default:
      retval = VMCI_ERROR_INVALID_VECTOR;
      break;
   }
   return retval;
}


/*
 *-----------------------------------------------------------------------------
 *
 * DatagramProcessNotify --
 *
 *      Callback to send a notificaton to a vmci process.
 *
 * Results:
 *      None
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

static void 
DatagramProcessNotify(void *notifyData,    // IN:
                      uint32 msgSize,      // IN:
                      VMCIDatagramMsg *msg)  // IN:
{
   VMCIDatagramProcess *dgmProc = (VMCIDatagramProcess *) notifyData;
   size_t dgmSize = VMCI_CALL_HEADERSIZE + msgSize;
   VMCICall *dgmProcCall = VMCI_AllocKernelMem(dgmSize, 
                                 VMCI_MEMORY_NONPAGED | VMCI_MEMORY_ATOMIC);
   if (!dgmProcCall) {
      return;
   }
   
   dgmProcCall->h.vector = VMCI_DATAGRAM_RECEIVE;
   dgmProcCall->h.size = dgmSize;
   memcpy(dgmProcCall->args, msg, msgSize);
   DatagramProcessSendCall(dgmProc, dgmProcCall, FALSE);
   /* dgmProcCall is freed when process reads call.. */
}


/*
 *-----------------------------------------------------------------------------
 *
 * VMCIDatagram_Init --
 *
 *      Register guest call handlers.
 *
 * Results:
 *      None
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

void 
VMCIDatagram_Init(void)
{
   int i;
   
   VMCI_InitLock(&hashTable.lock);
   for (i = 0; i < HASH_TABLE_SIZE; i++) {
      hashTable.entries[i] = NULL;
   }

   VMCI_RegisterHandler(VMCI_DATAGRAM_RECEIVE, DatagramReceive, 0, NULL);

   /* Register driver call handlers. */
   VMCI_RegisterDriverHandler(VMCI_DATAGRAM_CREATE_HND, DatagramDriverHandler);
   VMCI_RegisterDriverHandler(VMCI_DATAGRAM_DESTROY_HND, DatagramDriverHandler);
   VMCI_RegisterDriverHandler(VMCI_DATAGRAM_SEND, DatagramDriverHandler);
}


/*
 *-----------------------------------------------------------------------------
 *
 * VMCIDatagram_CheckHostCapabilities --
 *
 *      Verify that the host supports the hypercalls we need. If it does not,
 *      try to find fallback hypercalls and use those instead.
 *
 * Results:
 *      TRUE if required hypercalls (or fallback hypercalls) are
 *      supported by the host, FALSE otherwise.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

#define VMCI_DATAGRAM_NUM_VECTORS 3

Bool
VMCIDatagram_CheckHostCapabilities(void)
{
   int result;
   VMCIVectorsMsg *msg;
   uint32 callSize = sizeof(VMCIVectorsHdr) +
                     VMCI_DATAGRAM_NUM_VECTORS * sizeof(VMCI_Vector);
   VMCICall *checkCall = VMCI_AllocKernelMem(callSize, 0);

   if (checkCall == NULL) {
      VMCI_LOG(("VCMIDatagram: Check host: Insufficient memory.\n"));
      return FALSE;
   }

   checkCall->h.vector = VMCI_VECTORS_QUERY;
   checkCall->h.size = callSize;
   msg = (VMCIVectorsMsg *)checkCall->args;

   msg->numVectors = VMCI_DATAGRAM_NUM_VECTORS;
   msg->vectors[0] = VMCI_DATAGRAM_CREATE_HND;
   msg->vectors[1] = VMCI_DATAGRAM_DESTROY_HND;
   msg->vectors[2] = VMCI_DATAGRAM_SEND;

   result = VMCI_HyperCall(checkCall);
   VMCI_FreeKernelMem(checkCall);

   /* We need all 3 vectors. There are no fallbacks. */
   return (result == 0x7);
}


/*
 *----------------------------------------------------------------------
 *
 * VMCIDatagramProcess_Create --
 *
 *      Creates a new VMCIDatagramProcess object.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
VMCIDatagramProcess_Create(VMCIDatagramProcess **outDgmProc, // IN:
                           const int eventHnd)               // IN:
{
   VMCIDatagramProcess *dgmProc;

   dgmProc = VMCI_AllocKernelMem(sizeof *dgmProc, VMCI_MEMORY_NONPAGED);
   if (dgmProc == NULL) {
      return VMCI_ERROR_NO_MEM;
   }

   VMCI_InitLock(&dgmProc->callQueueLock);
   dgmProc->handle = VMCI_INVALID_HANDLE;
   VMCIHost_InitContext(&dgmProc->host, eventHnd);
   dgmProc->pendingCalls = 0;
   dgmProc->callQueue = NULL;

   *outDgmProc = dgmProc;
   return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * VMCIDatagramProcess_Destroy --
 *
 *      Destroys a VMCIDatagramProcess object.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

void
VMCIDatagramProcess_Destroy(VMCIDatagramProcess *dgmProc) // IN:
{
   ListItem *curr;
   ListItem *next;
   struct CallEntry *pcEntry;
   VMCILockFlags flags;

   if (!dgmProc) {
      return;
   }

   if (dgmProc->handle != VMCI_INVALID_HANDLE) {
      VMCIDatagram_DestroyHnd(dgmProc->handle);
      dgmProc->handle = VMCI_INVALID_HANDLE;
   }

   /* Flush dgmProc's call queue. */
   VMCI_GrabLock(&dgmProc->callQueueLock, &flags);
   LIST_SCAN_SAFE(curr, next, dgmProc->callQueue) {
      pcEntry = LIST_CONTAINER(curr, struct CallEntry, listItem);
      LIST_DEL(curr, &dgmProc->callQueue);
      ASSERT(pcEntry && pcEntry->call);
      VMCI_FreeKernelMem(pcEntry->call);
      VMCI_FreeKernelMem(pcEntry);
   }
   VMCI_ReleaseLock(&dgmProc->callQueueLock, flags);

   VMCI_FreeKernelMem(dgmProc);
}

/*
 *----------------------------------------------------------------------
 *
 * DatagramProcessSendCall --
 *
 *      Queues a VMCI datagram call for the appropriate target 
 *      process.
 *
 * Results:
 *      0 on success, appropriate error code otherwise.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static int
DatagramProcessSendCall(VMCIDatagramProcess *dgmProc, // IN: Target device handle
                        VMCICall *dgmProcCall,   // IN:
                        Bool mustCopy)                // IN: TRUE if must copy
{
   struct CallEntry *pcEntry;
   VMCILockFlags flags;

   ASSERT(dgmProc);
   if (!dgmProcCall || dgmProcCall->h.size < VMCI_CALL_HEADERSIZE) {
      return VMCI_ERROR_INVALID_ARGS;
   }

   /* Allocate process call entry and add it to the target process' queue. */
   pcEntry = VMCI_AllocKernelMem(sizeof *pcEntry,
                                 VMCI_MEMORY_NONPAGED | VMCI_MEMORY_ATOMIC);
   if (pcEntry == NULL) {
      return VMCI_ERROR_NO_MEM;
   }
   if (mustCopy) {
      pcEntry->call = VMCI_AllocKernelMem(dgmProcCall->h.size,
                                          VMCI_MEMORY_NONPAGED | 
                                          VMCI_MEMORY_ATOMIC);
      if (pcEntry->call == NULL) {
         VMCI_FreeKernelMem(pcEntry);
         return VMCI_ERROR_NO_MEM;
      }
      memcpy(pcEntry->call, dgmProcCall, dgmProcCall->h.size);
   } else {
      pcEntry->call = dgmProcCall;
   }

   VMCI_GrabLock(&dgmProc->callQueueLock, &flags);
   if (dgmProc->pendingCalls >= VMCI_MAX_NUM_CALLS) {
      if (mustCopy) {
         VMCI_FreeKernelMem(pcEntry->call);
      }
      VMCI_FreeKernelMem(pcEntry);
      VMCI_ReleaseLock(&dgmProc->callQueueLock, flags);
      return VMCI_ERROR_NO_MEM;
   }

   LIST_QUEUE(&pcEntry->listItem, &dgmProc->callQueue);
   dgmProc->pendingCalls++;
   VMCIHost_SignalCall(&dgmProc->host);
   VMCI_ReleaseLock(&dgmProc->callQueueLock, flags);

   return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * VMCIDatagramProcess_ReadCall --
 *
 *      Dequeues the next guest call and returns it to user level.
 *
 * Results:
 *      0 on success, appropriate error code otherwise.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
VMCIDatagramProcess_ReadCall(VMCIDatagramProcess *dgmProc, // IN:
                             void *buf,                    // OUT: User buffer
                             size_t bufSize)               // IN: size of buf
{
   struct CallEntry *pcEntry;
   ListItem *listItem;
   int result = VMCI_SUCCESS;
   VMCILockFlags flags;

   ASSERT(dgmProc);

   /* Dequeue the next dgmProc call entry. */
   VMCI_GrabLock(&dgmProc->callQueueLock, &flags);

   if (dgmProc->pendingCalls == 0) {
      VMCIHost_ClearCall(&dgmProc->host);
      VMCI_ReleaseLock(&dgmProc->callQueueLock, flags);
      return VMCI_ERROR_NO_MORE_CALLS;
   }

   listItem = LIST_FIRST(dgmProc->callQueue);
   ASSERT (listItem != NULL);

   pcEntry = LIST_CONTAINER(listItem, struct CallEntry, listItem);
   ASSERT(pcEntry->call);

   /* Check size of the userland buffer. */
   if (bufSize < pcEntry->call->h.size) {
      VMCI_ReleaseLock(&dgmProc->callQueueLock, flags);
      VMCI_LOG(("Userland buffer is too small.\n"));
      return VMCI_ERROR_NO_MEM;
   }
   
   LIST_DEL(listItem, &dgmProc->callQueue);
   dgmProc->pendingCalls--;
   if (dgmProc->pendingCalls == 0) {
      VMCIHost_ClearCall(&dgmProc->host);
   }
   VMCI_ReleaseLock(&dgmProc->callQueueLock, flags);

   if (VMCI_CopyToUser((void *)buf, pcEntry->call,
                       pcEntry->call->h.size) != 0) {
      result = VMCI_ERROR_NO_MEM;
   }
   VMCI_FreeKernelMem(pcEntry->call);
   VMCI_FreeKernelMem(pcEntry);

   return result;
}

