/* **********************************************************
 * Copyright 2006 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/

/*
 * vmciProcess.c --
 *
 *     VMCI Process code for guest driver.
 */

#ifdef linux
#include "driver-config.h"

#include <linux/kernel.h>
#include <linux/pci.h>

#endif // linux

#include "vmciInt.h"
#include "vmci_defs.h"
#include "vmci_kernel_defs.h"
#include "vmciProcess.h"
#include "vmciDatagram.h"
#include "vmci_infrastructure.h"
#include "circList.h"
#include "vmciUtil.h"

static ListItem *processList = NULL;
static VMCILock processLock;

/*
 *----------------------------------------------------------------------
 *
 * VMCIProcess_Init --
 *
 *      General init code.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

void
VMCIProcess_Init(void)
{
   VMCI_InitLock(&processLock);
}


/*
 *-----------------------------------------------------------------------------
 *
 * VMCIProcess_CheckHostCapabilities --
 *
 *      Verify that the host supports the hypercalls we need. If it does not,
 *      try to find fallback hypercalls and use those instead.
 *
 * Results:
 *      TRUE if required hypercalls (or fallback hypercalls) are
 *      supported by the host, FALSE otherwise.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

Bool
VMCIProcess_CheckHostCapabilities(void)
{
   /* VMCIProcess does not require any hypercalls. */
   return TRUE;
}


/*
 *----------------------------------------------------------------------
 *
 * VMCIProcess_Create --
 *
 *      Creates a new VMCI process.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
VMCIProcess_Create(VMCIProcess **outProcess,     // IN
                   int eventHnd)                 // IN
{
   VMCIProcess *process;
   VMCILockFlags flags;

   process = VMCI_AllocKernelMem(sizeof *process, VMCI_MEMORY_NONPAGED);
   if (process == NULL) {
      return VMCI_ERROR_NO_MEM;
   }

   process->pid = (VMCIId)(uintptr_t)process >> 1;

   process->sharedmemHandles = VMCIHandleArray_Create(0);
   if (process->sharedmemHandles == NULL) {
      VMCIHandleArray_Destroy(process->sharedmemHandles); 
      VMCI_FreeKernelMem(process);
      return VMCI_ERROR_NO_MEM;                                                           
   }      
   
   VMCI_GrabLock(&processLock, &flags);
   LIST_QUEUE(&process->listItem, &processList);
   VMCI_ReleaseLock(&processLock, flags);

   *outProcess = process;
   return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * VMCIProcess_Destroy --
 *
 *      Destroys a VMCI process.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

void
VMCIProcess_Destroy(VMCIProcess *process)
{
   VMCIHandle handle;
   VMCILockFlags flags;

   /* Cleanup the shared mem regions still open by process. */
   ASSERT(process->sharedmemHandles != NULL);
   handle = VMCIHandleArray_RemoveTail(process->sharedmemHandles);
   while (handle != VMCI_INVALID_HANDLE) {
      // XXX This should be fixed when sharedmem code refactoring is done.
#ifdef linux
      vmcisharedmem_destroy_handle(handle);
#else
      VMCISharedMemDestroyHandle(handle);
#endif // linux
      handle = VMCIHandleArray_RemoveTail(process->sharedmemHandles);
   }

   /* Dequeue process. */
   VMCI_GrabLock(&processLock, &flags);
   LIST_DEL(&process->listItem, &processList);
   VMCI_ReleaseLock(&processLock, flags);

   VMCIHandleArray_Destroy(process->sharedmemHandles);
   VMCI_FreeKernelMem(process);
}


/*
 *----------------------------------------------------------------------
 *
 * VMCIProcess_Get --
 *
 *      Get the process corresponding to the pid.
 *
 * Results:
 *      VMCI process on success, NULL otherwise.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

VMCIProcess *
VMCIProcess_Get(VMCIId processID)  // IN
{
   VMCIProcess *process = NULL;  
   ListItem *next;
   VMCILockFlags flags;

   VMCI_GrabLock(&processLock, &flags);
   if (processList == NULL) {
      goto out;
   }

   LIST_SCAN(next, processList) {
      process = LIST_CONTAINER(next, VMCIProcess, listItem);
      if (process->pid == processID) {
         break;
      }
   }

out:
   VMCI_ReleaseLock(&processLock, flags);
   return (process && process->pid == processID) ? process : NULL;
}

