/* ************************************************************************
 * Copyright 2006 VMware, Inc.  All rights reserved. -- VMware Confidential
 * ************************************************************************/

/* 
 * vmciResource.c --
 *
 *      Resource Access Control API for the Linux guest driver
 */
   
#if defined(linux)
#include "driver-config.h"
   
#include <linux/kernel.h>  
#include <linux/pci.h>
#elif defined(WIN32)
#include <ntddk.h>
#endif // linux

#include "vm_basic_types.h"
#include "vm_assert.h"
#include "vmci_defs.h"
#include "vmciUtil.h"
#include "vmci_infrastructure.h"
#include "vmciInt.h"
#include "vmci_kernel_defs.h"

#define LGPFX "VMCIResource: "

static int64 ResourceDriverHandler(void *ctObj,
                                   VMCIObjType ctObjType,
                                   VMCICall *call);

/*
 *-----------------------------------------------------------------------------
 *
 * VMCIResource_Init --
 *
 *      Register driver call handlers.
 *
 * Results:
 *      VMCI_SUCCESS if succesful, error if not.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

void
VMCIResource_Init(void)
{
   /* Register driver call handlers. */
   VMCI_RegisterDriverHandler(VMCI_RESOURCE_ADD_CLIENT_PRIVS,
                              ResourceDriverHandler);
   VMCI_RegisterDriverHandler(VMCI_RESOURCE_REMOVE_CLIENT_PRIVS,
                              ResourceDriverHandler);
   VMCI_RegisterDriverHandler(VMCI_RESOURCE_REMOVE_ALL_CLIENT_PRIVS,
                              ResourceDriverHandler);
}


/*
 *-----------------------------------------------------------------------------
 *
 * VMCIResource_CheckHostCapabilities --
 *
 *      Verify that the host supports the hypercalls we need. If it does not,
 *      try to find fallback hypercalls and use those instead.
 *
 * Results:
 *      TRUE if required hypercalls (or fallback hypercalls) are
 *      supported by the host, FALSE otherwise.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

#define VMCI_RESOURCE_NUM_VECTORS 3

Bool
VMCIResource_CheckHostCapabilities(void)
{
   int result;
   VMCIVectorsMsg *msg;
   uint32 callSize = sizeof(VMCIVectorsHdr) +
                     VMCI_RESOURCE_NUM_VECTORS * sizeof(VMCI_Vector);
   VMCICall *checkCall = VMCI_AllocKernelMem(callSize, 0);

   if (checkCall == NULL) {
      VMCI_LOG((LGPFX"Check host: Insufficient memory.\n"));
      return FALSE;
   }

   checkCall->h.vector = VMCI_VECTORS_QUERY;
   checkCall->h.size = callSize;
   msg = (VMCIVectorsMsg *)checkCall->args;

   msg->numVectors = VMCI_RESOURCE_NUM_VECTORS;
   msg->vectors[0] = VMCI_RESOURCE_ADD_CLIENT_PRIVS;
   msg->vectors[1] = VMCI_RESOURCE_REMOVE_CLIENT_PRIVS;
   msg->vectors[2] = VMCI_RESOURCE_REMOVE_ALL_CLIENT_PRIVS;

   result = VMCI_HyperCall(checkCall);
   VMCI_FreeKernelMem(checkCall);

   /* We need all 3 vectors. There are no fallbacks. */
   return (result == 0x7);
}


/*
 *-----------------------------------------------------------------------------
 *
 * VMCIResource_AddClientPrivileges --
 *
 *      Adds the given privileges to the client for the given resource.
 *
 * Results:
 *      VMCI_SUCCESS if succesful, error if not.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

int
VMCIResource_AddClientPrivileges(VMCIHandle resourceHandle,
                                 VMCIHandle clientHandle,
                                 int numAllowPrivs,
                                 VMCIResourcePrivilegeType *allowPrivs,
                                 int numDenyPrivs,
                                 VMCIResourcePrivilegeType *denyPrivs)
{
   uint32 retval = VMCI_SUCCESS;
   VMCIResourceAddClientPrivsHdr *addCall;
   uint32 callSize = sizeof *addCall +
      (numAllowPrivs + numDenyPrivs) * sizeof(VMCIResourcePrivilegeType);

   if (numAllowPrivs + numDenyPrivs < 1) {
      retval = VMCI_ERROR_INVALID_ARGS;
      goto done;
   }

   addCall = VMCI_AllocKernelMem(callSize, VMCI_MEMORY_NONPAGED);
   if (addCall == NULL) {
      VMCI_LOG(("Couldn't allocate memory for handle 0x%"FMT64"x.\n",
                resourceHandle));
      retval = VMCI_ERROR_NO_MEM;
      goto done;
   }

   addCall->hdr.vector = VMCI_RESOURCE_ADD_CLIENT_PRIVS;
   addCall->hdr.size = callSize;
   addCall->resourceHandle = resourceHandle;
   addCall->clientHandle = clientHandle;
   addCall->numAllowPrivs = numAllowPrivs;
   if (numAllowPrivs > 0) {
      memcpy((char *)addCall + sizeof *addCall , allowPrivs, 
	     numAllowPrivs * sizeof(VMCIResourcePrivilegeType));
   }
   addCall->numDenyPrivs = numDenyPrivs;
   if (numDenyPrivs > 0) {
      int offset =
	 sizeof *addCall + numAllowPrivs * sizeof(VMCIResourcePrivilegeType);
      memcpy((char *)addCall + offset, denyPrivs, 
	     numDenyPrivs * sizeof(VMCIResourcePrivilegeType));
   }

   retval = VMCI_HyperCall((VMCICall *)addCall);
   VMCI_FreeKernelMem(addCall);

  done:
   return retval;
}


/*
 *-----------------------------------------------------------------------------
 *
 * VMCIResource_RemoveClientPrivileges --
 *
 *      Removes the given privileges of the client for the given resource.
 *
 * Results:
 *      VMCI_SUCCESS if succesful, error if not.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

int
VMCIResource_RemoveClientPrivileges(VMCIHandle resourceHandle,
                                    VMCIHandle clientHandle,
                                    int numPrivs,
                                    VMCIResourcePrivilegeType *privs)
{
   uint32 retval = VMCI_SUCCESS;
   VMCIResourceRemoveClientPrivsHdr *removeCall;
   uint32 callSize = sizeof *removeCall + 
      numPrivs * sizeof(VMCIResourcePrivilegeType);

   if (numPrivs < 1) {
      retval = VMCI_ERROR_INVALID_ARGS;
      goto done;
   }

   removeCall = VMCI_AllocKernelMem(callSize, VMCI_MEMORY_NONPAGED);
   if (removeCall == NULL) {
      VMCI_LOG(("Couldn't allocate memory for handle 0x%"FMT64"x.\n",
                resourceHandle));
      retval = VMCI_ERROR_NO_MEM;
      goto done;
   }

   removeCall->hdr.vector = VMCI_RESOURCE_REMOVE_CLIENT_PRIVS;
   removeCall->hdr.size = callSize;
   removeCall->resourceHandle = resourceHandle;
   removeCall->clientHandle = clientHandle;
   removeCall->numPrivs = numPrivs;
   memcpy((char *)removeCall + sizeof *removeCall, privs, 
	  numPrivs * sizeof(VMCIResourcePrivilegeType));

   retval = VMCI_HyperCall((VMCICall *)removeCall);
   VMCI_FreeKernelMem(removeCall);

  done:
   return retval;
}


/*
 *-----------------------------------------------------------------------------
 *
 * VMCIResource_RemoveAllClientPrivileges --
 *
 *      Removes all the client's privileges to the given resource.
 *
 * Results:
 *      VMCI_SUCCESS if succesful, error if not.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

int
VMCIResource_RemoveAllClientPrivileges(VMCIHandle resourceHandle,
                                       VMCIHandle clientHandle)
{
   VMCIResourceRemoveAllClientPrivsHdr removeAllCall = 
      {{VMCI_RESOURCE_REMOVE_ALL_CLIENT_PRIVS, sizeof removeAllCall}, 
       resourceHandle, clientHandle};

   return VMCI_HyperCall((VMCICall *)&removeAllCall);
}


/*
 *------------------------------------------------------------------------------
 *
 * ResourceDriverHandler --
 *
 *      Parses the driver call and calls the appropriate function.
 *
 * Results:
 *      Value of called function.
 *
 * Side effects:
 *      None.
 *
 *------------------------------------------------------------------------------
 */

static int64
ResourceDriverHandler(void *ctObj,         // IN:
                      VMCIObjType ctObjType, // IN:
                      VMCICall *call) // IN:
{
   int retval;

   switch (call->h.vector) {
   case VMCI_RESOURCE_ADD_CLIENT_PRIVS:
   case VMCI_RESOURCE_REMOVE_CLIENT_PRIVS:
   case VMCI_RESOURCE_REMOVE_ALL_CLIENT_PRIVS:
      /* XXX Validate resource belongs to process. */
      retval = VMCI_HyperCall(call);
      break;
   default:
      retval = VMCI_ERROR_INVALID_VECTOR;
      break;
   }
   return retval;
}
