/* ************************************************************************
 * Copyright 2006 VMware, Inc.  All rights reserved. -- VMware Confidential
 * ************************************************************************/

/* 
 * vmciSharedMemory.c --
 *
 *      Shared memory API for the Linux guest driver.
 */
   
#include "driver-config.h"
   
#include <linux/kernel.h>  
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <asm/uaccess.h>

#include "compat_page.h"
#include "vm_atomic.h"
#include "vm_basic_types.h"
#include "vm_basic_defs.h"
#include "vm_assert.h"
#include "vmci_defs.h"
#include "vmci_kernel_defs.h"
#include "vmci_infrastructure.h"
#include "vmciInt.h"
#include "vmciProcess.h"
#include "vmciUtil.h"
#include "dbllnklst.h"
#include "pgtbl.h"
#include "iocontrols.h"

#define LGPFX "VMCISharedMem: "

typedef enum vmci_caller_context {
   VMCI_CONTEXT_USERLEVEL,
   VMCI_CONTEXT_KERNEL
} vmci_caller_context;

typedef struct vmcisharedmem_entry {
   DblLnkLst_Links    links;  
   VMCIHandle           handle;
   uint32             size;         /* Size of shared memory region in bytes. */
   VA                 va;           /* Only tracks kernel virtual addresses currently. */
   vmci_caller_context  is_user_va; /* User level or kernel virtual address? */
   char               page_file_name[VMCI_PATH_MAX];
   int                ref_count;
   PPN                *ppn_list;
} vmcisharedmem_entry;

typedef struct vmcisharedmem_state {
   DblLnkLst_Links list;        /* List head of shared memory regions. */
   spinlock_t lock;
} vmcisharedmem_state;


/* Local functions. */
static int vmcisharedmem_register(vmcisharedmem_entry *entry);
static Bool vmcisharedmem_unregister(VMCIHandle handle,
                                     vmci_caller_context is_user_va, 
                                     vmcisharedmem_entry **entry);
static Bool vmcisharedmem_is_registered(VMCIHandle handle);
static int vmcisharedmem_query(VMCIHandle handle);
static int vmcisharedmem_get_reference(VMCIHandle handle,
                                       vmci_caller_context is_user_va, VA *va,
                                       uint32 *size, char *page_file_name);
static int vmcisharedmem_create_work(uint32 size, VMCIHandle handle, 
                                     Bool create, VA *va);
static int vmcisharedmem_create_phys(PPN *ppns, uint32 size, VMCIHandle handle,
                                     Bool create);
static int vmcisharedmem_detach_work(VMCIHandle handle,
                                     vmci_caller_context is_user_va);
static VMCIHandle vmcisharedmem_create_user(VMCIProcess *process,
                                          char *page_file_name, VA va,
                                          uint32 size, VMCIHandle *handle);
static int vmcisharedmem_attach_user(VMCIProcess *process, VMCIHandle handle,
                                     VA va, char *page_file_name, uint32 size);
static VMCIHandle vmcisharedmem_create_user_work(VMCIProcess *process,
                                               char *page_file_name, VA va,
                                               uint32 size, VMCIHandle handle, 
                                               Bool create);
static int vmcisharedmem_detach_user(VMCIProcess *process, VMCIHandle handle);
static int vmcisharedmem_query_user(VMCIHandle handle, uint32 *size_in_bytes, 
                                    char *page_file_name);

DEBUG_ONLY(static void vmcisharedmem_dump_state(void);)


static vmcisharedmem_state shared_mem;


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_init --
 *
 *      Initalizes shared memory API.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

void 
vmcisharedmem_init(void)
{
   spin_lock_init(&shared_mem.lock);
   DblLnkLst_Init(&shared_mem.list);
}


/*
 *-----------------------------------------------------------------------------
 *
 * VMCISharedMem_CheckHostCapabilities --
 *
 *      Verify that the host supports the hypercalls we need. If it does not,
 *      try to find fallback hypercalls and use those instead.
 *
 * Results:
 *      TRUE if required hypercalls (or fallback hypercalls) are
 *      supported by the host, FALSE otherwise.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

#define VMCI_SHAREDMEM_NUM_VECTORS 4

Bool
VMCISharedMem_CheckHostCapabilities(void)
{
   int result;
   VMCIVectorsMsg *msg;
   uint32 callSize = sizeof(VMCIVectorsHdr) +
                     VMCI_SHAREDMEM_NUM_VECTORS * sizeof(VMCI_Vector);
   VMCICall *checkCall = VMCI_AllocKernelMem(callSize, 0);

   if (checkCall == NULL) {
      VMCI_LOG((LGPFX"Check host: Insufficient memory.\n"));
      return FALSE;
   }

   checkCall->h.vector = VMCI_VECTORS_QUERY;
   checkCall->h.size = callSize;
   msg = (VMCIVectorsMsg *)checkCall->args;

   msg->numVectors = VMCI_SHAREDMEM_NUM_VECTORS;
   msg->vectors[0] = VMCI_SHAREDMEM_CREATE;
   msg->vectors[1] = VMCI_SHAREDMEM_ATTACH;
   msg->vectors[2] = VMCI_SHAREDMEM_DETACH;
   msg->vectors[3] = VMCI_SHAREDMEM_QUERY;

   result = VMCI_HyperCall(checkCall);
   VMCI_FreeKernelMem(checkCall);

   /* We need all 4 vectors. There are no fallbacks. */
   return (result == 0xf);
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_entry_create --
 *
 *      Creates an object to store information about a given shared memory 
 *      region.
 *
 * Results:
 *      An initialized shared memory entry on success, or NULL on failure.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

vmcisharedmem_entry * 
vmcisharedmem_entry_create(uint32 mem_size)   // IN: Size of shared memory region
{
   vmcisharedmem_entry *entry;

   entry = kmalloc(sizeof *entry, GFP_KERNEL);
   if (entry == NULL) {
      printk(KERN_ALERT LGPFX "Failed to allocate memory for sharedmem entry\n");
      return NULL;
   }

   entry->ppn_list = kmalloc(CEILING(mem_size, PAGE_SIZE) * sizeof *entry->ppn_list,
                             GFP_KERNEL);
   if (entry->ppn_list == NULL) {
      printk(KERN_ALERT LGPFX "Failed to allocate memory for ppn list in entry\n");
      kfree(entry);
      return NULL;
   }

   entry->size = mem_size;
   memset(entry->page_file_name, 0, sizeof entry->page_file_name);
   entry->handle = VMCI_INVALID_HANDLE;
   entry->ref_count = 1;
   entry->va = 0;

   return entry;
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_entry_destroy --
 *
 *      Frees memory for a shared memory entry.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

void 
vmcisharedmem_entry_destroy(vmcisharedmem_entry * entry)   // IN
{
   if (entry != NULL) {
      kfree(entry->ppn_list);
      kfree(entry);
   }   
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_entry_dup --
 *
 *      Create a shared memory entry and copies an existing entry into it.
 *
 * Results:
 *      Copied shared memory entry on success, or NULL on failure.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

vmcisharedmem_entry *
vmcisharedmem_entry_dup(vmcisharedmem_entry * entry)   // IN
{
   vmcisharedmem_entry * new_entry;
   size_t ppn_list_size;

   if (entry == NULL) {
      return NULL;
   }   
   
   new_entry = vmcisharedmem_entry_create(entry->size);
   if (new_entry == NULL) {
      printk(KERN_ALERT LGPFX "Failed to allocate duplicate entry\n");
      return NULL;
   }

   ppn_list_size = CEILING(entry->size, PAGE_SIZE) * sizeof *entry->ppn_list;
   new_entry->ppn_list = kmalloc(ppn_list_size, GFP_KERNEL);
   if (new_entry->ppn_list == NULL) {
      vmcisharedmem_entry_destroy(new_entry);
      printk(KERN_ALERT LGPFX "Failed to allocate memory for ppn list in entry\n");
      return NULL;
   }

   /* Copy all the fields over. */
   new_entry->handle = entry->handle;
   new_entry->size = entry->size;
   new_entry->va = entry->va;
   new_entry->is_user_va = entry->is_user_va;
   new_entry->ref_count = entry->ref_count;

   memcpy(new_entry->ppn_list, entry->ppn_list, ppn_list_size);
   memcpy(new_entry->page_file_name, entry->page_file_name, 
          sizeof entry->page_file_name);

   /* The new entry is not on any list. */
   DblLnkLst_Init(&new_entry->links);
   return new_entry;
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_register --
 *
 *      Registers information about the given shared memory region.
 *
 * Results:
 *      VMCI_SUCCESS on success, appropriate error code otherwise.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

int 
vmcisharedmem_register(vmcisharedmem_entry *entry)                   // IN
{
   
   ASSERT(entry != NULL);
   ASSERT(entry->ppn_list != NULL);
   ASSERT(entry->handle != VMCI_INVALID_HANDLE);
   ASSERT(entry->size > 0);
   ASSERT(entry->is_user_va == VMCI_CONTEXT_KERNEL || 
          entry->is_user_va == VMCI_CONTEXT_USERLEVEL);

   spin_lock(&shared_mem.lock);

   if (vmcisharedmem_is_registered(entry->handle)) {
      printk(KERN_INFO LGPFX "Duplicate handle\n");
      spin_unlock(&shared_mem.lock);
      return VMCI_ERROR_DUPLICATE_ENTRY;
   }

   DblLnkLst_Init(&entry->links);
   DblLnkLst_LinkLast(&shared_mem.list, &entry->links);

   spin_unlock(&shared_mem.lock);
   return VMCI_SUCCESS;
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_unregister --
 *
 *      Unregisters the given shared memory region.
 *
 * Results:
 *      A copy of the relevant shared memory entry on success, NULL otherwise.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

Bool 
vmcisharedmem_unregister(VMCIHandle handle,                 // IN
                         vmci_caller_context is_user_va,  // IN
                         vmcisharedmem_entry **mem_entry) // OUT: copy of entry
{
   DblLnkLst_Links *link;
   vmcisharedmem_entry *entry;
   Bool found = FALSE;

   spin_lock(&shared_mem.lock);
   
   for (link = shared_mem.list.next; link != &shared_mem.list; link = link->next) {
      entry = DblLnkLst_Container(link, vmcisharedmem_entry, links);
      if (entry->handle != handle) {
         continue;
      }

      if (is_user_va != entry->is_user_va) {
         printk(KERN_ALERT LGPFX "Detaching shared memory from incorrect"
                  "context.\n");
         break;
      }

      found = TRUE;
      entry->ref_count--;
      if (mem_entry != NULL) {
         *mem_entry = vmcisharedmem_entry_dup(entry);
      }

      if (entry->ref_count == 0) {
         DblLnkLst_Unlink1(&entry->links);
         vmcisharedmem_entry_destroy(entry);
      }
      break;
   }
   spin_unlock(&shared_mem.lock);
   return found;
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_is_registered --
 *
 *      Determines if a given shared memory region is already registered.
 *      Assumes caller holds appropriate locks.
 *
 * Results:
 *      TRUE if already registered in this guest, FALSE otherwise.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

Bool 
vmcisharedmem_is_registered(VMCIHandle handle)
{
   DblLnkLst_Links *link;

   for (link = shared_mem.list.next; link != &shared_mem.list; link = link->next) {
      vmcisharedmem_entry *entry;

      entry = DblLnkLst_Container(link, vmcisharedmem_entry, links);
      if (entry->handle == handle) {
         return TRUE;
         break;
      }
   }
   return FALSE;
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_get_reference --
 *
 *      Gets a reference to a shared memory region already created or 
 *      attached to in this VM.
 *
 * Results:
 *      VMCI_SUCCESS or appropriate error code.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

int 
vmcisharedmem_get_reference(VMCIHandle handle,                // IN: 
                            vmci_caller_context is_user_va, // IN: 
                            VA *va,                         // OUT: Start VA for region
                            uint32 *size,                   // OUT: Size in bytes
                            char *page_file_name)           // OUT:
{
   DblLnkLst_Links *link;
   int result = VMCI_ERROR_NO_HANDLE;

   ASSERT(size != NULL);

   spin_lock(&shared_mem.lock);
   for (link = shared_mem.list.next; link != &shared_mem.list; link = link->next) {
      vmcisharedmem_entry *entry;

      entry = DblLnkLst_Container(link, vmcisharedmem_entry, links);
      if (entry->handle == handle) {
         /* 
          * XXX: Kernel mode and user mode can't currently share the same 
          * memory within a VM till we implement mmap to associate arbitrary
          * PPNs with a given VA range. 
          */

         if (is_user_va != entry->is_user_va) {
            printk(KERN_INFO LGPFX "Cannot share same memory region "
                                    "from different calling contexts");
            result = VMCI_SHAREDMEM_ERROR_BAD_CONTEXT;
            break;
         }

         entry->ref_count++;
         if (size != NULL) {
            *size = entry->size;
         }
         if (va != NULL) {
            *va = entry->va;
         }
         if (page_file_name != NULL) {
            memcpy(page_file_name, entry->page_file_name, sizeof entry->page_file_name);
         }
         result = VMCI_SUCCESS;
         break;
      }
   }

   spin_unlock(&shared_mem.lock);
   return result;
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_query --
 *
 *      Get the size in bytes of a given shared memory region, if it exists.
 *
 * Results:
 *      Returns size of the region on success, error value otherwise.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

int
vmcisharedmem_query(VMCIHandle handle)
{
   VMCISharedMemQueryHdr query_call = 
      {{VMCI_SHAREDMEM_QUERY, sizeof query_call}, handle};

   return VMCI_HyperCall((VMCICall *)&query_call);
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_create_phys --
 *
 *      Creates shared memory region using the given physical pages
 *      and returns a handle to it. Shared memory is always created
 *      on a page-granularity.
 *
 *      XXX: With some more housekeeping, this could be exported
 *           to other modules to enable sharing physical pages.
 *
 * Results:
 *      Returns handle on success, negative error value otherwise.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

int 
vmcisharedmem_create_phys(PPN *ppns,          // IN
                          uint32 size,        // IN: 
                          VMCIHandle handle,    // IN
                          Bool create)        // IN: create or attach
{
   int result = VMCI_SUCCESS;
   int i;
   int call_size;
   int num_pages;
   VMCISharedMemCreateHdr *create_call;

   if (ppns == NULL) {
      return VMCI_ERROR_INVALID_ARGS;
   }

   /* We always share an integral number of pages. */
   num_pages = CEILING(size, PAGE_SIZE);
   call_size = sizeof *create_call + num_pages * sizeof(PPN);
   create_call = kmalloc(call_size, GFP_ATOMIC);
   if (create_call == NULL) {
      printk(KERN_ERR LGPFX "Failed to allocate shared memory hypercall\n");
      result = VMCI_ERROR_NO_MEM;
      goto out;
   }

   /* Create shared memory on the host. */
   create_call->hdr.vector = create ? VMCI_SHAREDMEM_CREATE :
                                      VMCI_SHAREDMEM_ATTACH;
   create_call->hdr.size = call_size;
   create_call->handle = handle;
   create_call->memSize = size;
   
   /* PPNs are placed after struct. */
   memcpy((char *)create_call + sizeof *create_call, ppns, 
	  num_pages * sizeof ppns[0]);

   result = VMCI_HyperCall((VMCICall *)create_call);
   kfree(create_call);
   if (result != VMCI_SUCCESS) {
      printk(KERN_ERR "Failed to share pages, err 0x%x.\n", result);
      goto out;
   }

   /* Pin all shared PPNs. */
   for (i = 0; i < num_pages; i++) {
      struct page *page;
      page = pfn_to_page(ppns[i]);
      get_page(page);
   }

out:
   return result;
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_create_work --
 *
 *      Creates or attaches to a shared memory region and returns a handle 
 *      to it. Also returns the kernel logical address of the  start of the 
 *      shared memory region.
 *
 * Results:
 *      Returns handle on success, negative error value otherwise.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

int 
vmcisharedmem_create_work(uint32 size,       // IN: Size of shared region in bytes
                          VMCIHandle handle,   // IN:
                          Bool create,       // IN: Create or attach
                          VA *va)            // OUT: Linear address of start of region
{
   struct page *page;
   vmcisharedmem_entry *entry = NULL;
   int i;
   unsigned num_pages = CEILING(size, PAGE_SIZE);
   int result = VMCI_ERROR_GENERIC;
   
   ASSERT(va != NULL);

   page = alloc_pages(GFP_KERNEL, compat_get_order(ROUNDUP(size, PAGE_SIZE)));
   if (page == NULL) {
      printk(KERN_ERR LGPFX "Error allocating pages to share\n");
      return VMCI_ERROR_NO_MEM;
   }

   entry = vmcisharedmem_entry_create(size);
   if (entry == NULL) {
      result = VMCI_ERROR_NO_MEM;
      goto error;
   }

   /* alloc_pages() returns contiguous physical pages. Create PPN list. */
   entry->ppn_list[0] = page_to_pfn(page);
   for (i = 1; i < num_pages; i++) {
      entry->ppn_list[i] = entry->ppn_list[0] + i;
   }

   /* Register entry. */
   entry->va = (VA)page_address(page);
   entry->handle = handle;
   entry->is_user_va = VMCI_CONTEXT_KERNEL;
   result = vmcisharedmem_register(entry);
   if (result != VMCI_SUCCESS) {
      printk(KERN_ALERT LGPFX "Failed to create shared memory state\n");
      vmcisharedmem_entry_destroy(entry);
      goto error;
   }

   result = vmcisharedmem_create_phys(entry->ppn_list, size, handle, create);
   if (result != VMCI_SUCCESS) {
      /* entry freed in vmcisharedmem_unregister */
      vmcisharedmem_unregister(handle, VMCI_CONTEXT_KERNEL, NULL);
      printk(KERN_ERR LGPFX "Error sharing memory\n");
      goto error;
   }
   
   *va = entry->va;
   return VMCI_SUCCESS;

error: 

   __free_pages(page, compat_get_order(ROUNDUP(size, PAGE_SIZE)));
   return result;
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_create --
 *
 *      Kernel interface to create a shared memory region. Creates region and 
 *      returns a handle to it. Also returns the kernel logical address of the
 *      start of the shared memory region.
 *
 * Results:
 *      Returns handle on success, negative error value otherwise.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

VMCIHandle 
vmcisharedmem_create(size_t size,    // IN: Size of shared region in bytes
                     VA *va)   
{
   VMCIHandle handle;

   if (va == NULL) {
      printk(KERN_INFO LGPFX "Invalid arguments. Cannot return linear address\n");
      return VMCI_INVALID_HANDLE;
   }

   handle = VMCI_CreateHandle();
   if (handle == VMCI_INVALID_HANDLE) {
      return VMCI_INVALID_HANDLE;
   }

   if (vmcisharedmem_create_work(size, handle, TRUE, va) == VMCI_SUCCESS) {
      return handle;
   } else {
      return VMCI_INVALID_HANDLE;
   }
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_attach --
 *
 *      Kernel interface to attach to a shared memory region. 
 *
 * Results:
 *      Returns VMCI_SUCCESS on success, negative error value otherwise.
 *      Also returns the kernel logical address of the
 *      start of the shared memory region.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

int
vmcisharedmem_attach(VMCIHandle handle,       // IN
                     VA *va,                // OUT
                     uint32 *size_in_bytes) // OUT: 
{
   int32 size;
   int result;

   /* Return locally if this shared memory region has already been attached to. */
   result = vmcisharedmem_get_reference(handle, VMCI_CONTEXT_KERNEL, 
                                        va, &size, NULL);
   if (result == VMCI_SUCCESS) {
      if (size_in_bytes != NULL) {
         *size_in_bytes = size;
      }
      return result;
   }

   size = vmcisharedmem_query(handle);
   if (size <= 0) {
      printk(KERN_INFO LGPFX "Error attaching to shared memory region. "
             "Handle (%"FMT64"d) not found.\n", handle);
      return VMCI_ERROR_NO_HANDLE;
   }

   result = vmcisharedmem_create_work(size, handle, FALSE, va);
   if (result == VMCI_SUCCESS && size_in_bytes != NULL) {
      *size_in_bytes = size;
   }

   return result;
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_detach_work --
 *
 *      Detach from the given shared memory region.
 *
 * Results:
 *      VMCI_SUCCESS on successfully detaching from the shared memory region, 
 *      appropriate error code otherwise.
 *
 * Side effects:
 *      Guest PPNs that were part of this shared memory region are unpinned.
 *      Shared memory region is destroyed on the host if this is the last VM
 *      to detach from it.
 *      
 *
 *-----------------------------------------------------------------------------
 */

int
vmcisharedmem_detach_work(VMCIHandle handle,                // IN:
                          vmci_caller_context is_user_va) // IN:
{
   uint32 num_pages;
   int i;
   int result = VMCI_SUCCESS;
   vmcisharedmem_entry *entry = NULL;
   VMCISharedMemDetachHdr detach_call =
      {{VMCI_SHAREDMEM_DETACH, sizeof detach_call}, handle};

   if (!vmcisharedmem_unregister(handle, is_user_va, &entry)) {
      return VMCI_ERROR_NO_HANDLE;
   }

   if (entry == NULL) {
      /* Handle was found, we should have an entry. */
      return VMCI_ERROR_NO_MEM;
   }

   if (entry->ref_count > 0) {
      /* Other clients are currently attached to this region as well. */
      goto out;
   }

   result = VMCI_HyperCall((VMCICall *)&detach_call);
   if (result != VMCI_SUCCESS) {
      printk(KERN_ERR LGPFX "Error detaching from shared memory region %"FMT64"d\n", 
             handle);
   }

   /* Unpin shared pages. */
   num_pages = CEILING(entry->size, PAGE_SIZE);
   ASSERT(entry->ppn_list != NULL);
   for (i = 0; i < num_pages; i++) {
      struct page *next = pfn_to_page(entry->ppn_list[i]);
      put_page(next);
   }

   if (entry->is_user_va == VMCI_CONTEXT_KERNEL) {
      /* Free pages as well. */

      struct page *page = virt_to_page(entry->va);
      if (page != NULL) {      
         __free_pages(page, compat_get_order(ROUNDUP(entry->size, PAGE_SIZE)));
      }
   }

out: 
   if (result == VMCI_SUCCESS && entry->ref_count == 0) {
      result = VMCI_SUCCESS_LAST_DETACH;
   }
   vmcisharedmem_entry_destroy(entry);;
   return result;
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_detach --
 *
 *      Kernel API to detach from the given shared memory region.
 *
 * Results:
 *      TRUE on successfully detaching from the shared memory region, 
 *      FALSE otherwise.
 *
 * Side effects:
 *      Guest PPNs that were part of this shared memory region are unpinned.
 *      Shared memory region is destroyed on the host if this is the last VM
 *      to detach from it.
 *      
 *
 *-----------------------------------------------------------------------------
 */

int
vmcisharedmem_detach(VMCIHandle handle)
{
   return vmcisharedmem_detach_work(handle, VMCI_CONTEXT_KERNEL);
}


#ifdef VMX86_DEBUG

/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_dump_state --
 *
 *      Dumps current shared memory state for debugging.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

void 
vmcisharedmem_dump_state(void)
{
   DblLnkLst_Links *link;
   vmcisharedmem_entry *entry;
   int count = 1;

   printk(KERN_DEBUG LGPFX "Dumping shared memory state:\n");
   spin_lock(&shared_mem.lock);
   
   for (link = shared_mem.list.next; link != &shared_mem.list; link = link->next) {
      entry = DblLnkLst_Container(link, vmcisharedmem_entry, links);
      printk(KERN_DEBUG LGPFX "%d. Handle %"FMT64"d, %d pages, start VA =%"FMTVA"x (%s), "
             "ref count =%d.\n", count, entry->handle, 
             (uint32)CEILING(entry->size, PAGE_SIZE), entry->va, 
             entry->is_user_va == VMCI_CONTEXT_KERNEL ? "kernel" : "user level",
             entry->ref_count);
      count++;
   }
   spin_unlock(&shared_mem.lock);
   printk(KERN_DEBUG LGPFX "End of shared memory state\n");
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_test_registration --
 *
 *      Test registering/unregistering shared memory regions.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

void 
vmcisharedmem_test_registration(void)
{
   int i;
   vmcisharedmem_entry *entry = vmcisharedmem_entry_create(PAGE_SIZE);
   if (entry == NULL) {
      return;
   }

   /* Test registering/unregistering shared memory regions. */
   for (i = 0; i < 10; i++) {
      entry->handle = i;
      entry->is_user_va = VMCI_CONTEXT_KERNEL;
      vmcisharedmem_register(entry);
   }

   vmcisharedmem_dump_state();

   for (i = 0; i < 10; i++) {
      vmcisharedmem_unregister((VMCIHandle)i, VMCI_CONTEXT_KERNEL, NULL);
   }
   vmcisharedmem_dump_state();
   vmcisharedmem_entry_destroy(entry);

}

#endif // VMX86_DEBUG


/* User level shared memory API handlers. */

/*
 *----------------------------------------------------------------------
 *
 * vmcisharedmem_ioctl_handler --
 *
 *     Handles requests to create, attach, detach and query shared
 *     memory regions.
 *     
 * Results:
 *      VMCI_SUCCESS on success, appropriate error code otherwise.
 *
 * Side effects:
 *      None.
 *
 *---------------------------------------------------------------------- 
 */

int
vmcisharedmem_ioctl_handler(VMCIProcess *process, // IN
                            unsigned int iocmd,   // IN
                            void *ioarg)          // IN/OUT
{
   int result = VMCI_SUCCESS;
   VMCISharedMemInfo *memInfo;

   ASSERT(process != NULL);

   memInfo = kmalloc(sizeof *memInfo, GFP_ATOMIC);
   if (memInfo == NULL) {
      printk("Failed to allocate memory to get ioctl parameters\n");
      return VMCI_ERROR_NO_MEM;
   }

   result = copy_from_user(memInfo, (char*)ioarg, sizeof *memInfo);
   if (result != 0) {
      printk("Failed to get ioctl parameters for shared memory\n");
      result = VMCI_ERROR_INVALID_ARGS;
      goto out;
   }

   switch (iocmd) { 
   case IOCTLCMD_VMCI_SHAREDMEM_CREATE:
      result = vmcisharedmem_create_user(process, memInfo->pageFileName, (VA)memInfo->va,
                                         memInfo->size, &memInfo->handle);
      break;

   case IOCTLCMD_VMCI_SHAREDMEM_ATTACH:
      result = vmcisharedmem_attach_user(process, memInfo->handle, (VA)memInfo->va, 
                                         memInfo->pageFileName, memInfo->size);
      goto out;

   case IOCTLCMD_VMCI_SHAREDMEM_QUERY:
      result = vmcisharedmem_query_user(memInfo->handle, &memInfo->size, 
                                        memInfo->pageFileName);
      break;

   case IOCTLCMD_VMCI_SHAREDMEM_DETACH:
      result = vmcisharedmem_detach_user(process, memInfo->handle);
      break;

   default:
      NOT_REACHED();
      result = VMCI_ERROR_INVALID_ARGS;
      break;
   }

   memInfo->result = result;
   if (result >= VMCI_SUCCESS) {
      result = copy_to_user((char *)ioarg, memInfo, sizeof *memInfo);
   }

out:
   kfree(memInfo);
   return result;
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_create_user --
 *
 *      Handler for user level request to create a shared memory region. 
 *      Creates and returns a new handle.
 *
 * Results:
 *      VMCI_SUCCESS on success, appropriate error code otherwise.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

VMCIHandle 
vmcisharedmem_create_user(VMCIProcess *process,  // IN
                          char *page_file_name,  // IN
                          VA va,                 // IN
                          uint32 size,           // IN: size of shared region in bytes
                          VMCIHandle *handle)      // IN/OUT:
{
   VMCIHandle hnd;
   int result;

   ASSERT(handle != NULL);

   hnd = (*handle == VMCI_INVALID_HANDLE) ? VMCI_CreateHandle() : *handle;
   if (hnd == VMCI_INVALID_HANDLE) {
      printk(KERN_ALERT "Failed to create handle\n");
      return VMCI_ERROR_NO_HANDLE;
   }

   if (VMCIHandleArray_HasEntry(process->sharedmemHandles, hnd)) {
      printk(KERN_ALERT LGPFX "Process has already attached to region %"FMT64"d\n",
             hnd);
      return VMCI_ERROR_REGION_ALREADY_SHARED;
   }

   result =  vmcisharedmem_create_user_work(process, page_file_name, 
                                            va, size, hnd, TRUE);
   if (result == VMCI_SUCCESS) {
      *handle = hnd;
   }
   return result;
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_create_user_work --
 *
 *      Create or attach to a shared memory region starting at a user
 *      level virtual address. 
 *
 * Results:
 *      VMCI_SUCCESS on success, appropriate error code otherwise.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */


VMCIHandle 
vmcisharedmem_create_user_work(VMCIProcess *process,   // IN
                               char *page_file_name,   // IN
                               VA va,                  // IN
                               uint32 size,            // IN: size of region in bytes
                               VMCIHandle handle,        // IN
                               Bool create)            // IN: create or attach?
{
   int i;
   int result;
   uint32 num_pages;
   vmcisharedmem_entry *entry;

   if ((va & (PAGE_SIZE - 1)) != 0) {
      printk(KERN_ALERT "Cannot share memory that is not page-aligned "
	     "(%"FMTVA"x)\n", va);
      return VMCI_ERROR_NOT_PAGE_ALIGNED;
   }
   
   entry = vmcisharedmem_entry_create(size);
   if (entry == NULL) {
      printk(KERN_ALERT "Failed to create entry\n");
      return VMCI_ERROR_NO_MEM;
   }

   entry->va = va;
   entry->is_user_va = VMCI_CONTEXT_USERLEVEL;
   entry->handle = handle;

   /* Get PPNs corresponding to the user level virtual address. */
   num_pages = CEILING(size, PAGE_SIZE);
   for (i = 0; i < num_pages; i++) {
      entry->ppn_list[i] = PgtblVa2MPN(va + PAGE_SIZE * i);
      if (entry->ppn_list[i] == INVALID_MPN) {
         printk(KERN_ALERT LGPFX "Failed to get PPN for va %p\n", 
                (void *)va + PAGE_SIZE * i);
         result = VMCI_ERROR_INVALID_ARGS;
         goto error;
      }
   }

   memcpy(entry->page_file_name, page_file_name, sizeof entry->page_file_name);
   result = vmcisharedmem_register(entry);
   if (result != VMCI_SUCCESS) {
      printk(KERN_ALERT LGPFX "Failed to create shared memory state\n");
      goto error;
   }

   result = vmcisharedmem_create_phys(entry->ppn_list, size, handle, create);
   if (result != VMCI_SUCCESS) {
      vmcisharedmem_unregister(entry->handle, VMCI_CONTEXT_USERLEVEL, NULL);
      printk(KERN_ERR LGPFX "Error sharing memory\n");
      return result;
   }

   VMCIHandleArray_AppendEntry(&process->sharedmemHandles, handle);
   printk(KERN_DEBUG LGPFX "Created shared memory (%"FMT64"d)\n", handle);
   return VMCI_SUCCESS;

error:
   vmcisharedmem_entry_destroy(entry);
   return result;
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_attach_user --
 *
 *      Handler for user level request to attach to a shared memory region. 
 *      Attaches to region and returns a handle to it. 
 *
 * Results:
 *      VMCI_SUCCESS on success, error code otherwise.
 * 
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

int 
vmcisharedmem_attach_user(VMCIProcess *process,   // IN
                          VMCIHandle handle,        // IN
                          VA va,                  // IN
                          char *page_file_name,   // IN
                          uint32 size_in_bytes)   // IN
{
   int result;
   uint32 size;
   char filename[VMCI_PATH_MAX];

   ASSERT(process);

   if (VMCIHandleArray_HasEntry(process->sharedmemHandles, handle)) {
      printk(KERN_ALERT LGPFX "Process has already attached to region "
	     "%"FMT64"d\n", handle);
      return VMCI_ERROR_REGION_ALREADY_SHARED;
   }

   /* 
    * Return locally if this shared memory region has already been 
    * created/attached to by other process.
    */
   result = vmcisharedmem_get_reference(handle, VMCI_CONTEXT_USERLEVEL, 
                                        NULL, &size, filename);
   if (result == VMCI_SUCCESS) {
      if (size != size_in_bytes || 
         memcmp(filename, page_file_name, sizeof filename) != 0) {

         /* Release reference first. */
         vmcisharedmem_unregister(handle, VMCI_CONTEXT_USERLEVEL, NULL);
         return VMCI_ERROR_INVALID_ARGS;
      }

      /* Append handle to process's shared memory regions array. */
      VMCIHandleArray_AppendEntry(&process->sharedmemHandles, handle);

      return VMCI_SUCCESS;
   } else if (result != VMCI_ERROR_NO_HANDLE) {
      printk(KERN_INFO LGPFX "Error getting reference to existing entry\n");
      return result;
   }
   
   if (page_file_name[0] == '\0' || size_in_bytes == 0 || va == 0) {
      printk(KERN_ALERT LGPFX "Invalid parameters for attach request: "
             "size = %d, va = %"FMTVA"x, filename[0] = %c\n", size_in_bytes, 
             va, page_file_name[0]);
      return VMCI_ERROR_INVALID_ARGS;
   }   

   size = vmcisharedmem_query(handle);
   if (size <= 0) {
      printk(KERN_INFO LGPFX "Error attaching to shared memory region. "
            "Size for handle (%"FMT64"d) not found.\n", handle);
      return VMCI_ERROR_NO_HANDLE;
   }

   if (size != size_in_bytes) {
      printk(KERN_INFO LGPFX "Size mismatch for attach request"
             " for handle %"FMT64"d \n", handle);
      return VMCI_ERROR_INVALID_SIZE;
   }

   return vmcisharedmem_create_user_work(process, page_file_name, va, 
                                         size, handle, FALSE);
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_query_user --
 *
 *      Handler for user level request to get the size of a shared 
 *      memory region. Also returns the page file name if one exists.
 *
 * Results:
 *      VMCI_SUCCESS on success, error code otherwise.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

int 
vmcisharedmem_query_user(VMCIHandle handle,        // IN
                         uint32 *size_in_bytes,  // OUT
                         char *page_file_name)   // OUT
{
   int result = VMCI_ERROR_NO_HANDLE; 
   uint32 size;

   ASSERT(size_in_bytes != NULL);
   ASSERT(page_file_name != NULL);

   result = vmcisharedmem_get_reference(handle, VMCI_CONTEXT_USERLEVEL, 
                                        NULL, size_in_bytes, page_file_name);
   if (result == VMCI_SUCCESS) {
      /* Release the reference just obtained. */
      vmcisharedmem_unregister(handle, VMCI_CONTEXT_USERLEVEL, NULL);
      return result;
   }
                                      
   if (result == VMCI_ERROR_NO_HANDLE) {
      /* Query the host. */
      
      size = vmcisharedmem_query(handle);
      if (size > 0) {
         /* Region has not been attached to in this VM and doesnt have a page file. */
         memset(page_file_name, 0, VMCI_PATH_MAX);
         *size_in_bytes = size;
         result = VMCI_SUCCESS;
      } else {
         result = size;
      }
   }
   return result;
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_detach_user --
 *
 *       Handler for user level request to detach from the 
 *       given shared memory region.
 *
 * Results:
 *      VMCI_SUCCESS on success, error code otherwise. 
 *
 * Side effects:
 *      Shared memory region is destroyed on the host if this is the last VM
 *      to detach from it.
 *      
 *
 *-----------------------------------------------------------------------------
 */

int
vmcisharedmem_detach_user(VMCIProcess *process,  // IN
                          VMCIHandle handle)       // IN
{
   int result;

   if (!VMCIHandleArray_HasEntry(process->sharedmemHandles, handle)) {
      printk(KERN_ALERT LGPFX "Process does not own shared region %"FMT64"d\n", handle);
      return VMCI_ERROR_NO_ACCESS;
   }

   result = vmcisharedmem_detach_work(handle, VMCI_CONTEXT_USERLEVEL);
   if (result >= VMCI_SUCCESS) {
      VMCIHandleArray_RemoveEntry(process->sharedmemHandles, handle);
   }
   return result;
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmcisharedmem_destroy_handle --
 *
 *       Thin wrapper that detaches from the given user level shared 
 *       memory region.
 *
 * Results:
 *      VMCI_SUCCESS on success, error code otherwise. 
 *
 * Side effects:
 *      Shared memory region is destroyed on the host if this is the last VM
 *      to detach from it.
 *      
 *
 *-----------------------------------------------------------------------------
 */

int
vmcisharedmem_destroy_handle(VMCIHandle handle)     // IN
{
   return vmcisharedmem_detach_work(handle, VMCI_CONTEXT_USERLEVEL);
}
