/* **********************************************************
 * Copyright 2006 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/

/*
 * vmciUtil.c
 *
 * Small utility function for allocating kernel memory and copying data.
 *
 */

#if defined(linux)
#include "driver-config.h"
 #ifdef KERNEL_2_2 
   #include <linux/slab.h>
#else
   #include <linux/malloc.h>
#endif
#include "compat_wait.h"
#elif defined(_WIN32)
#ifndef WINNT_DDK
#error  This file only works with the NT ddk 
#endif // WINNT_DDK
#include <ntddk.h>
#else 
#error "platform not supported."
#endif //linux

#define LGPFX "VMCIUtil: "

#include "vmware.h"
#include "vmci_defs.h"
#include "vmci_kernel_defs.h"
#include "vmciInt.h"
#include "vmciProcess.h"
#include "vmciResource.h"
#include "vmciDatagram.h"
#include "vmciSharedMemory.h"
#include "vmciTest.h"


static int64 UtilDriverHandler(void *vmciObj,
                               VMCIObjType vmciObjType,
                               VMCICall *call);
static Bool UtilInformGuestCalls(void);

/*
 *-----------------------------------------------------------------------------
 *
 * VMCIUtil_Init --
 *
 *      Register driver call handlers.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

void
VMCIUtil_Init(void)
{
   VMCI_RegisterDriverHandler(VMCI_GET_CONTEXT_ID, UtilDriverHandler);
}


/*
 *-----------------------------------------------------------------------------
 *
 * VMCIUtil_CheckHostCapabilities --
 *
 *      Verify that the host supports the hypercalls we need. If it does not,
 *      try to find fallback hypercalls and use those instead.
 *
 * Results:
 *      TRUE if required hypercalls (or fallback hypercalls) are
 *      supported by the host, FALSE otherwise.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

#define VMCI_UTIL_NUM_VECTORS 1

Bool
VMCIUtil_CheckHostCapabilities(void)
{
   int result;
   VMCIVectorsMsg *msg;
   uint32 callSize = sizeof(VMCIVectorsHdr) +
                     VMCI_UTIL_NUM_VECTORS * sizeof(VMCI_Vector);
   VMCICall *checkCall = VMCI_AllocKernelMem(callSize, 0);

   if (checkCall == NULL) {
      VMCI_LOG((LGPFX"Check host: Insufficient memory.\n"));
      return FALSE;
   }

   checkCall->h.vector = VMCI_VECTORS_QUERY;
   checkCall->h.size = callSize;
   msg = (VMCIVectorsMsg *)checkCall->args;

   msg->numVectors = VMCI_UTIL_NUM_VECTORS;
   msg->vectors[0] = VMCI_GET_CONTEXT_ID;

   result = VMCI_HyperCall(checkCall);
   VMCI_FreeKernelMem(checkCall);

   /* We need the vector. There are no fallbacks. */
   return (result == 0x1);
}


/*
 *----------------------------------------------------------------------
 *
 * VMCI_CopyToUser
 *
 *      Copy memory to the user application from a  kernel buffer. 
 *
 * Results:
 *      0 if successful and VMCI_ERROR_NO_ACCESS on error.
 *
 * Side effects:
 *      None.
 *----------------------------------------------------------------------
 */

int 
VMCI_CopyToUser(void *dst, const void *src, unsigned int len)
{
#ifdef _WIN32
   __try {
      ProbeForWrite(dst, len, 1);
      RtlCopyMemory(dst, src, len);      
   } __except(EXCEPTION_EXECUTE_HANDLER) { 
      return STATUS_ACCESS_VIOLATION;
   }
   return 0;
#else // linux
   /*
    * copy_to_user() can deschedule this task on kernels where the low
    * latency patch has been applied. --hpreg
    */
   return copy_to_user(dst, src, len);
#endif // _WIN32
}


/*
 *-----------------------------------------------------------------------------
 *
 * VMCI_GetResourceID --
 *
 *      Return a context unique id. XXX Currently we just inc a number.
 *
 * Results:
 *      Unique resource id.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

VMCIId
VMCI_GetResourceID(void) 
{
   static VMCIId rid = 0;
   return ++rid;
}


/*
 *-----------------------------------------------------------------------------
 *
 * VMCI_GetContextID --
 *
 *      Returns the context id. First time it will make a hypercall to obtain
 *      it.
 *
 * Results:
 *      Context id.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

VMCIId
VMCI_GetContextID(void)
{
   static VMCIId cid = VMCI_INVALID_ID;

   if (cid == VMCI_INVALID_ID) {
      VMCICall get_cid_call =
	 {{VMCI_GET_CONTEXT_ID, VMCI_CALL_HEADERSIZE}, {0}};
      cid = VMCI_HyperCall(&get_cid_call);
   }
   return cid;
}


/*
 *-----------------------------------------------------------------------------
 *
 * VMCI_CreateHandle --
 *
 *      Create a new VMCI handle with a unique resource id.
 *
 * Results:
 *      VMCI handle.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

VMCIHandle
VMCI_CreateHandle(void)
{
   VMCIId context_id;
   VMCIId resource_id;
   /* Create a handle for shared memory region. */
   context_id = VMCI_GetContextID();
   if (context_id == VMCI_INVALID_ID) {
      VMCI_LOG((LGPFX "Failed to create handle\n"));
      return context_id;
   }

   resource_id = VMCI_GetResourceID();
   return VMCI_MAKE_HANDLE(context_id, resource_id);
}


/*
 *-----------------------------------------------------------------------------
 *
 * UtilDriverHandler --
 *
 *      Parses the driver call and calls the appropriate function.
 *
 * Results:
 *      Value of called function.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

static int64
UtilDriverHandler(void *vmciObj,           // IN:
                  VMCIObjType vmciObjType, // IN:
                  VMCICall *call)          // IN
{
   int64 retval;
   ASSERT(call);

   switch(call->h.vector) {
   case VMCI_GET_CONTEXT_ID:
      retval = VMCI_GetContextID();
      break;

   default:
      retval = VMCI_ERROR_INVALID_VECTOR;
      break;
   }
   
   return retval;
}


/*
 *-----------------------------------------------------------------------------
 *
 * UtilInformGuestCalls --
 *
 *      Tell the host which guestcalls are supported by this context, by
 *      issuing a VMCI_VECTORS_INFORM hypercall.
 *
 * Results:
 *      TRUE if the host accepts our guestcalls, FALSE otherwise.
 *
 * Side effects:
 *      Fallback mechanisms for backwards compatibility may be enabled in
 *      vmmon for this context.
 *
 *-----------------------------------------------------------------------------
 */

static Bool
UtilInformGuestCalls(void)
{
   VMCICall *call;
   VMCIVectorsMsg *msg;
   int i, numVectors, callSize, index, result;

   /* Count the number of registered guestcall vectors. */
   numVectors = 0;
   for (i = 0; i < VMCI_VECTOR_MAX; i++) {
      if (vmciVector[i].handler != NULL) {
         numVectors++;
      }
   }

   callSize = sizeof(VMCIVectorsHdr) + numVectors * sizeof(VMCI_Vector);
   call = VMCI_AllocKernelMem(callSize, 0);
   if (call == NULL) {
      VMCI_LOG((LGPFX"Unable to negotiate vectors: Insufficient memory.\n"));
      return FALSE;
   }

   call->h.vector = VMCI_VECTORS_INFORM;
   call->h.size = callSize;
   msg = (VMCIVectorsMsg *)call->args;

   /* Put all registered guestcall vectors in the msg->vectors array. */
   index = 0;
   for (i = 0; i < VMCI_VECTOR_MAX; i++) {
      if (vmciVector[i].handler != NULL) {
         msg->vectors[index++] = i;
      }
   }
   msg->numVectors = numVectors;
   ASSERT(index == numVectors);
   
   result = VMCI_HyperCall(call);
   VMCI_FreeKernelMem(call);
   if (result < VMCI_SUCCESS) {
      VMCI_LOG((LGPFX"Host declined our guestcalls (%d).\n",
                result));
      return FALSE;
   }

   return TRUE;
}


/*
 *-----------------------------------------------------------------------------
 *
 * VMCI_CheckHostCapabilities --
 *
 *      Tell host which guestcalls we support and let each API check
 *      that the host supports the hypercalls it needs. If a hypercall
 *      is not supported, the API can check for a fallback hypercall,
 *      or fail the check.
 *
 * Results:
 *      TRUE if successful, FALSE otherwise.
 *
 * Side effects:
 *      Fallback mechanisms may be enabled in the API and vmmon.
 *
 *-----------------------------------------------------------------------------
 */

Bool
VMCI_CheckHostCapabilities(void)
{
   Bool result = UtilInformGuestCalls();
   result &= VMCIProcess_CheckHostCapabilities();
   result &= VMCIResource_CheckHostCapabilities();
   result &= VMCIDatagram_CheckHostCapabilities();
   result &= VMCISharedMem_CheckHostCapabilities();
   result &= VMCIUtil_CheckHostCapabilities();

   VMCI_LOG((LGPFX"Host capability check: %s\n", result ? "PASSED" : "FAILED"));

   return result;
}

