/* ************************************************************************
 * Copyright 2005 VMware, Inc.  All rights reserved. -- VMware Confidential
 * ************************************************************************/

/* 
 * vmci.c --
 *
 *      Linux guest driver for the VMCI device.
 */
   
#include "driver-config.h"
   
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
   
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 9)
#include <linux/moduleparam.h>
#endif
   
#include "compat_pci.h"
#include "compat_wait.h"
#include "compat_init.h"
#include "compat_ioport.h"
#include "compat_interrupt.h"
#include "compat_page.h"
#include "vm_basic_types.h"
#include "vm_device_version.h"
#include "kernelStubs.h"
#include "iocontrols.h"
#include "vmci_defs.h"
#include "vmciInt.h"
#include "vmci_infrastructure.h"
#include "vmciDatagram.h"
#include "vmciResource.h"
#include "vmciProcess.h"
#include "vmciSharedMemory.h"
#include "vmciUtil.h"

#ifdef VMX86_DEBUG
#include "vmciTest.h"
#endif // VMX86_DEBUG

#define LGPFX "VMCI: "
#define VMCI_DEVICE_MINOR_NUM 0

typedef struct vmci_device {
   struct semaphore lock;

   unsigned int ioaddr;
   unsigned int ioaddr_size;
   unsigned int irq;

   Bool         enabled;
   spinlock_t   dev_spinlock;
} vmci_device;

static int vmci_probe_device(struct pci_dev *pdev,
                             const struct pci_device_id *id);
static void vmci_remove_device(struct pci_dev* pdev);
static int vmci_open(struct inode *inode, struct file *file);
static int vmci_close(struct inode *inode, struct file *file);
static int vmci_ioctl(struct inode *inode, struct file *file, 
                      unsigned int cmd, unsigned long arg);
static unsigned int vmci_poll(struct file *file, poll_table *wait);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 19)
static compat_irqreturn_t vmci_interrupt(int irq, void *dev_id, 
                                         struct pt_regs * regs);
#else
static compat_irqreturn_t vmci_interrupt(int irq, void *dev_id);
#endif
static void dispatch_guest_calls(unsigned long data);

static const struct pci_device_id vmci_ids[] = {
   { PCI_DEVICE(PCI_VENDOR_ID_VMWARE, PCI_DEVICE_ID_VMWARE_VMCI), },
   { 0 },
};

static struct file_operations vmci_ops = {
   .owner   = THIS_MODULE,
   .open    = vmci_open,
   .release = vmci_close,
   .ioctl   = vmci_ioctl,
   .poll    = vmci_poll,
};

static struct pci_driver vmci_driver = {
   .name     = "vmci",
   .id_table = vmci_ids,
   .probe = vmci_probe_device,
   .remove = vmci_remove_device,
};

static vmci_device vmci_dev;

/* We dynamically request the device major number at init time. */
static int device_major_nr = 0;

DECLARE_TASKLET(vmci_tasklet, dispatch_guest_calls, 
                (unsigned long)&vmci_dev);

/* 
 * Allocate a buffer for guest calls globally to avoid repeated 
 * allocation in the interrupt handler's atomic context. 
 */ 
static uint8 *guest_call_buffer = NULL;  
static uint32 guest_call_buffer_size = PAGE_SIZE;

/* Array of all the guestcall functions indexed by their vector. */
VMCIHandlerEntry vmciVector[VMCI_VECTOR_MAX];

/* Array of all the driver call functions indexed by their vector. */
VMCIDriverCallHandler driverCallHandler[VMCI_VECTOR_MAX];


/*
 *-----------------------------------------------------------------------------
 *
 * vmci_init --
 *
 *      Initialization, called by Linux when the module is loaded.
 *
 * Results:
 *      Returns 0 for success, negative errno value otherwise.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

static int
vmci_init(void)
{
   int err = -ENOMEM;

   printk(KERN_DEBUG "Inside vmci_init\n");

   /* Register device node ops. */
   err = register_chrdev(0, "vmci", &vmci_ops);
   if (err < 0) {
      printk(KERN_ERR "Unable to register vmci device\n"); 
      return err;
   }
   device_major_nr = err;

   printk("VMCI: Major device number is: %d\n", device_major_nr);

   /* Initialize device data. */
   init_MUTEX(&vmci_dev.lock);
   spin_lock_init(&vmci_dev.dev_spinlock);
   vmci_dev.enabled = FALSE;

   guest_call_buffer = kmalloc(guest_call_buffer_size, GFP_ATOMIC);
   if (guest_call_buffer == NULL) {
      goto error;
   }
  
   VMCIProcess_Init();
   VMCIResource_Init();
   VMCIDatagram_Init();
   vmcisharedmem_init();
   VMCIUtil_Init();

   DEBUG_ONLY(VMCITest_Init();)

   /* This should be last to make sure we are done initializing. */
   err = pci_register_driver(&vmci_driver);
   if (err < 0) {
      goto error;
   }

   return 0;

error:
   unregister_chrdev(device_major_nr, "vmci");
   kfree(guest_call_buffer);
   return err;
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmci_exit --
 *
 *      Cleanup, called by Linux when the module is unloaded.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

static void
vmci_exit(void)
{
   int err = 0;

   printk(KERN_INFO "Inside vmci_exit\n");
   pci_unregister_driver(&vmci_driver);
   
   err = unregister_chrdev(device_major_nr, "vmci");
   if (err < 0) {
      printk(KERN_ERR "Unable to register vmci device\n"); 
   }
   
   kfree(guest_call_buffer);
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmci_probe_device --
 *
 *      Most of the initialization at module load time is done here.
 *
 * Results:
 *      Returns 0 for success, an error otherwise.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

static int
vmci_probe_device(struct pci_dev *pdev,           // IN: vmci PCI device
                  const struct pci_device_id *id) // IN: matching device ID
{
   unsigned int ioaddr;
   unsigned int ioaddr_size;
   unsigned int capabilities;
   int result;

   printk(KERN_INFO "Probing for vmci/PCI.\n");

   result = compat_pci_enable_device(pdev);
   if (result) {
      printk(KERN_ERR "Cannot VMCI device %s: error %d\n",
             compat_pci_name(pdev), result);
      return result;
   }
   ioaddr = compat_pci_resource_start(pdev, 0);
   ioaddr_size = compat_pci_resource_len(pdev, 0);

   /*
    * Request I/O region with adjusted base address and size. The adjusted
    * values are needed and used if we release the region in case of failure.
    */

   if (!compat_request_region(ioaddr, ioaddr_size, "vmci")) {
      printk(KERN_INFO "vmci: Another driver already loaded "
                       "for device in slot %s.\n", compat_pci_name(pdev));
      goto pci_disable;
   }

   printk(KERN_INFO "Found vmci/PCI at %#x, irq %u.\n", ioaddr, pdev->irq);

   /*
    * Verify that the VMCI Device supports the capabilities that
    * we need. If the device is missing capabilities that we would
    * like to use, check for fallback capabilities and use those
    * instead (so we can run a new VM on old hosts). Fail the load if
    * a required capability is missing and there is no fallback.
    *
    * Right now, we need Hypercall and Guestcall. There are no
    * fallbacks.
    */
   capabilities = inl(ioaddr + VMCI_CAPS_ADDR);

   if ((capabilities & VMCI_CAPS_HYPERCALL) == 0) {
      printk(KERN_ERR "VMCI device does not support hypercalls.\n");
      goto release;
   }
   if ((capabilities & VMCI_CAPS_GUESTCALL) == 0) {
      printk(KERN_ERR "VMCI device does not support guestcalls.\n");
      goto release;
   }

   /* Let the host know which capabilities we intend to use. */
   outl(VMCI_CAPS_HYPERCALL | VMCI_CAPS_GUESTCALL, ioaddr + VMCI_CAPS_ADDR);

   /* Device struct initialization. */
   down(&vmci_dev.lock);
   if (vmci_dev.enabled) {
      printk(KERN_ERR "VMCI device already enabled.\n");
      goto unlock;
   }

   vmci_dev.ioaddr = ioaddr;
   vmci_dev.ioaddr_size = ioaddr_size;
   vmci_dev.irq = pdev->irq;

   /* Check host hypercalls. */
   if (!VMCI_CheckHostCapabilities()) {
      goto unlock;
   }

   /* Enable device. */
   vmci_dev.enabled = TRUE;
   pci_set_drvdata(pdev, &vmci_dev);

   if (request_irq(vmci_dev.irq, vmci_interrupt, COMPAT_IRQF_SHARED, 
                   "vmci", &vmci_dev)) {
      printk(KERN_ERR "vmci: irq %u in use\n", vmci_dev.irq);
      goto unlock;
   }

   printk(KERN_INFO "Registered vmci device.\n");

   up(&vmci_dev.lock);

   /* Enable interrupts. */
   outl(VMCI_CONTROL_INT_ENABLE, vmci_dev.ioaddr + VMCI_CONTROL_ADDR);

   return 0;

 unlock:
   up(&vmci_dev.lock);
 release:
   release_region(ioaddr, ioaddr_size);
 pci_disable:
   compat_pci_disable_device(pdev);
   return -EBUSY;
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmci_remove_device --
 *
 *      Cleanup, called for each device on unload.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

static void
vmci_remove_device(struct pci_dev* pdev)
{
   struct vmci_device *dev = pci_get_drvdata(pdev);

   printk(KERN_INFO "Removing vmci device\n");

   down(&dev->lock);
   printk(KERN_INFO "Resetting vmci device\n");
   outl(VMCI_CONTROL_RESET, vmci_dev.ioaddr + VMCI_CONTROL_ADDR);
   free_irq(dev->irq, dev);
   release_region(dev->ioaddr, dev->ioaddr_size);
   dev->enabled = FALSE;

   printk(KERN_INFO "Unregistered vmci device.\n");
   up(&dev->lock);
   
   compat_pci_disable_device(pdev);
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmci_open --
 *
 *      Open device.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

static int
vmci_open(struct inode *inode,  // IN
          struct file *file)    // IN
{
   VMCIGuestDeviceHandle *devHndl;
   int errcode;

   printk(KERN_INFO "Opening vmci device\n");

   if (MINOR(inode->i_rdev) != VMCI_DEVICE_MINOR_NUM) {
      return -ENODEV;
   }

   down(&vmci_dev.lock);
   if (!vmci_dev.enabled) {
      printk(KERN_INFO "Received open on uninitialized vmci device.\n");
      errcode = -ENODEV;
      goto unlock;
   }

   /* Do open ... */
   devHndl = VMCI_AllocKernelMem(sizeof *devHndl, VMCI_MEMORY_NONPAGED);
   if (!devHndl) {
      printk(KERN_INFO "Failed to create device obj when opening device.\n");
      errcode = -ENOMEM;
      goto unlock;
   }
   devHndl->obj = NULL;
   devHndl->objType = VMCIOBJ_NOT_SET;
   file->private_data = devHndl;

   up(&vmci_dev.lock);

   return 0;

 unlock:
   up(&vmci_dev.lock);
   return errcode;
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmci_close --
 *
 *      Close device.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

static int
vmci_close(struct inode *inode,  // IN
           struct file *file)    // IN
{
   VMCIGuestDeviceHandle *devHndl =
      (VMCIGuestDeviceHandle *) file->private_data;

   if (devHndl) {
      if (devHndl->objType == VMCIOBJ_PROCESS) {
         VMCIProcess_Destroy((VMCIProcess *) devHndl->obj);
      } else if (devHndl->objType == VMCIOBJ_DATAGRAM_PROCESS) {
         VMCIDatagramProcess_Destroy((VMCIDatagramProcess *) devHndl->obj);
      }
      VMCI_FreeKernelMem(devHndl);
      file->private_data = NULL;
   }
   return 0;
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmci_ioctl --
 *
 *      IOCTL interface to device.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

static int
vmci_ioctl(struct inode *inode,  // IN
           struct file *file,    // IN
           unsigned int cmd,     // IN
           unsigned long arg)    // IN
{
   int retval;
   VMCIGuestDeviceHandle *devHndl =
      (VMCIGuestDeviceHandle *) file->private_data;

   if (devHndl == NULL) {
      return -EINVAL;
   }

   switch (cmd) {
   case IOCTLCMD_VMCI_CREATE_PROCESS: {
      if (devHndl->objType != VMCIOBJ_NOT_SET) {
         printk("VMCI: Received IOCTLCMD_VMCI_CREATE_PROCESS on "
                "initialized handle.\n");
         retval = -EINVAL;
         break;
      }
      ASSERT(!devHndl->obj);
      retval = VMCIProcess_Create((VMCIProcess **) &devHndl->obj, -1);
      if (retval != 0) {
         printk("VMCI: Failed to create process.\n");
         break;
      }
      devHndl->objType = VMCIOBJ_PROCESS;
      break;
   }

   case IOCTLCMD_VMCI_CREATE_DATAGRAM_PROCESS: {
      if (devHndl->objType != VMCIOBJ_NOT_SET) {
         printk("VMCI: Received "
                "IOCTLCMD_VMCI_CREATE_DATAGRAM_PROCESS on "
                "initialized handle.\n");
         retval = -EINVAL;
         break;
      }
      ASSERT(!devHndl->obj);
      retval = VMCIDatagramProcess_Create((VMCIDatagramProcess **) &devHndl->obj,
                                                               -1);
      if (retval != 0) {
         printk("VMCI: Failed to create datagram process.\n");
         break;
      }
      devHndl->objType = VMCIOBJ_DATAGRAM_PROCESS;
      break;
   }

   case IOCTLCMD_VMCI_HYPERCALL: {
      VMCIHypercallInfo callInfo;
      VMCICall *hyperCall;

      if (devHndl->objType != VMCIOBJ_PROCESS &&
            devHndl->objType != VMCIOBJ_DATAGRAM_PROCESS) {
         retval = -EINVAL;
         break;
      }

      /* Get size of the hypercall data. */
      retval = copy_from_user(&callInfo, (char *)arg, sizeof callInfo);
      if (retval != 0) {
         printk(KERN_DEBUG "vmci_ioctl(): Error getting hypercall data %d\n", 
		retval);
         retval = -EFAULT;
         break;
      }

      hyperCall = kmalloc(callInfo.len, GFP_KERNEL);
      if (hyperCall == NULL) {
         printk(KERN_DEBUG "vmci_ioctl(): Cannot allocate memory to dispatch "
		"hypercall\n");
         retval = -ENOMEM;
         break;
      }

      /* Now get the hypercall command and arguments. */
      retval = copy_from_user(hyperCall, (char *)(VA)callInfo.addr, 
			      callInfo.len);
      if (retval != 0) {
         printk(KERN_DEBUG "vmci_ioctl(): Error getting hypercall arguments: %d\n",
		retval);
         retval = -EFAULT;
         break;
      }

      printk(KERN_DEBUG "VMCI: Hypercall vector %d, size %d.\n", 
	     hyperCall->h.vector, hyperCall->h.size);

      ASSERT(devHndl->obj);
      callInfo.result.val64 =
         VMCI_DispatchDriverCall(devHndl->obj, devHndl->objType, hyperCall);
      printk(KERN_DEBUG "VMCI: Hypercall result 0x%"FMT64"x.\n",
	     callInfo.result.val64);
      kfree(hyperCall);
      retval = copy_to_user((void *)arg, &callInfo, sizeof callInfo);
      break;
   }
   
   case IOCTLCMD_VMCI_READ_CALL: {
      VMCIReadCallInfo readCallInfo;
      void *bufferUVA;
      size_t size;

      if (devHndl->objType != VMCIOBJ_DATAGRAM_PROCESS) {
         retval = -EINVAL;
         break;
      }

      retval = copy_from_user(&readCallInfo, (void *) arg,
                              sizeof readCallInfo);
      if (retval) {
         printk(KERN_NOTICE "VMCI: copy_from_user failed.\n");
         retval = -EFAULT;
         break;
      }

      bufferUVA = (void *) ((uintptr_t) readCallInfo.bufferUVA);
      size = readCallInfo.size;

      retval = VMCIDatagramProcess_ReadCall((VMCIDatagramProcess *) devHndl->obj,
                                             bufferUVA, size);
      break;
   }

   case IOCTLCMD_VMCI_SHAREDMEM_CREATE: 
   case IOCTLCMD_VMCI_SHAREDMEM_ATTACH: 
   case IOCTLCMD_VMCI_SHAREDMEM_QUERY: 
   case IOCTLCMD_VMCI_SHAREDMEM_DETACH: 
      if (devHndl->objType != VMCIOBJ_PROCESS) {
         retval = -EINVAL;
         break;
      }
      retval = vmcisharedmem_ioctl_handler((VMCIProcess *) devHndl->obj, cmd,
                                           (void *) arg);
      break;

   default:
      printk(KERN_DEBUG "vmci_ioctl(): unknown ioctl 0x%x.\n", cmd);
      retval = -EINVAL;
      break;
   }

   return retval;
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmci_poll --
 *
 *      vmci poll function
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

static unsigned int
vmci_poll(struct file *file, // IN
          poll_table *wait)  // IN
{
   VMCILockFlags flags;
   unsigned int mask = 0;
   VMCIGuestDeviceHandle *devHndl =
      (VMCIGuestDeviceHandle *) file->private_data;
   VMCIDatagramProcess *dgmProc;

   /* 
    * Check for call to this VMCI process. 
    */
   
   if (!devHndl || devHndl->objType != VMCIOBJ_DATAGRAM_PROCESS) {
      return mask;
   }
   dgmProc = (VMCIDatagramProcess *) devHndl->obj;
   ASSERT(dgmProc);

   if (wait != NULL) {
      poll_wait(file, &dgmProc->host.waitQueue, wait);
   }

   VMCI_GrabLock(&dgmProc->callQueueLock, &flags);
   if (dgmProc->pendingCalls > 0) {
      mask = POLLIN;
   }
   VMCI_ReleaseLock(&dgmProc->callQueueLock, flags);

   return mask;
}


/*
 *-----------------------------------------------------------------------------
 *
 * vmci_interrupt --
 *
 *      Interrupt handler.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 19)
static compat_irqreturn_t
vmci_interrupt(int irq,               // IN
               void *clientdata,      // IN
               struct pt_regs *regs)  // IN
#else
static compat_irqreturn_t
vmci_interrupt(int irq,               // IN
               void *clientdata)      // IN
#endif
{
   vmci_device *dev = clientdata;
   unsigned int icr = 0;

   if (dev == NULL) {
      printk (KERN_DEBUG "vmci_interrupt(): irq %d for unknown device.\n",
              irq);
      return COMPAT_IRQ_NONE;
   }

   /* Acknowledge interrupt and determine what needs doing. */
   icr = inl(dev->ioaddr + VMCI_ICR_ADDR);
   
   switch (icr) {
      case VMCI_ICR_GUESTCALL:
         tasklet_schedule(&vmci_tasklet);
         break;
      default: 
         if (icr) {
            printk(KERN_INFO LGPFX"Interrupt cause (%d) unknown\n", icr);
         }
         return COMPAT_IRQ_NONE;
   }

   return COMPAT_IRQ_HANDLED;
}


/*
 *-----------------------------------------------------------------------------
 *
 * VMCI_HyperCall --
 *
 *      VM to hypervisor call mechanism. We use the standard VMware naming
 *      convention since shared code is calling this function as well.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

int
VMCI_HyperCall(VMCICall *hypercall)
{
   int retval = 0;
   unsigned long flags;

   /* Check args. */
   if (hypercall == NULL || hypercall->h.vector >= VMCI_VECTOR_MAX) {
      return VMCI_ERROR_INVALID_VECTOR;
   }
   
   /*
    * Need to acquire spinlock on the device because
    * the hypercall data may be spread over multiple pages and the monitor may
    * interleave device user rpc calls from multiple VCPUs. Acquiring the
    * spinlock precludes that possibility. Disabling interrupts to avoid
    * getting guest calls during a "rep out" and possibly landing up in this
    * function.
    */
   spin_lock_irqsave(&vmci_dev.dev_spinlock, flags);

   /* 
    * Do the hypercall and retrieve the return value from ebx, we add ebx to
    * the clobber list since it is used to return the result of the hypercall in.
    */
   __asm__ __volatile__(
      "cld\n\t"
      "rep outsb\n\t"
      "movl %%ebx, %0\n\t"
      : "=r"(retval)
      : "d"(vmci_dev.ioaddr + VMCI_HYPERCALL_ADDR),
	"c"(hypercall->h.size), "S"(hypercall)
      : "%ebx"
      );

   spin_unlock_irqrestore(&vmci_dev.dev_spinlock, flags);

   return retval;
}


/*
 *-----------------------------------------------------------------------------
 *
 * dispatch_guest_calls --
 *
 *      Reads and executes all available guest calls.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      Reads data from the device.
 *
 *-----------------------------------------------------------------------------
 */

void
dispatch_guest_calls(unsigned long data)
{
   VMCICall *call;
   vmci_device *dev = (vmci_device *)data;
   uint32 guestcall_port;

   if (dev == NULL) {
      printk(KERN_DEBUG "vmci: dispatch_guest_calls(): no vmci device"
	     "present.\n");
      return;
   }
      
   guestcall_port = dev->ioaddr + VMCI_GUESTCALL_ADDR;
   
   if (guest_call_buffer == NULL) {
      printk(KERN_DEBUG "vmci: dispatch_guest_calls(): no buffer present.\n");
      return;
   }

   insb(guestcall_port, guest_call_buffer, guest_call_buffer_size);
   call = (VMCICall *)guest_call_buffer; 

   while (call->h.vector != VMCI_ERROR_INVALID_VECTOR) {
      uint32 remainingBytes;

      if (call->h.vector < VMCI_VECTOR_MAX) {
         VMCI_Dispatch(VMCI_HOST_CONTEXT_ID, call);

         /* On to the next call. */
         call = (VMCICall *)((uint8 *)call + 
				  VMCI_GuestCallGetSize(call));
      } else if (call->h.vector != VMCI_ERROR_MORE_DATA) {
         /* Skip invalid guest call. */
         printk(KERN_INFO "Invalid guest call vector (%d)\n", call->h.vector);
         call = (VMCICall *)((uint8 *)call +
				  VMCI_GuestCallGetSize(call));
      }

      remainingBytes =
	 guest_call_buffer + guest_call_buffer_size - (uint8 *)call;

      if (remainingBytes >= VMCI_CALL_HEADERSIZE && 
          call->h.vector == VMCI_ERROR_MORE_DATA && 
          VMCI_GuestCallGetSize(call) > guest_call_buffer_size) {
      
         /* A guest call larger than the current buffer needs to be retrieved. */
         guest_call_buffer_size = VMCI_GuestCallGetSize(call);
         kfree(guest_call_buffer);
         guest_call_buffer = kmalloc(guest_call_buffer_size, GFP_ATOMIC);
	 if (guest_call_buffer == NULL) {
	    printk(KERN_DEBUG "Failed to allocate buffer of size %d.\n",
		   guest_call_buffer_size);
	    return;
	 }
      }

      if ((remainingBytes < VMCI_CALL_HEADERSIZE) || 
          call->h.vector == VMCI_ERROR_MORE_DATA) {

         /* Get the next batch of guest calls. */
         insb(guestcall_port, guest_call_buffer, guest_call_buffer_size);
         call = (VMCICall *)guest_call_buffer; 
      } 
   }
}


module_init(vmci_init);
module_exit(vmci_exit);
MODULE_DEVICE_TABLE(pci, vmci_ids);
